import numpy as np
import pandas as pd
from pandas import DataFrame
from statsmodels.tsa.stattools import grangercausalitytests


def calc_xcorr(target: str, ingested_data: DataFrame, max_lags: int, modes: [str] = ["pearson"]) -> dict:
    """
    Calculate the cross-correlation for the `ingested data`.
    Use `target` time-series column as target; the correlation is computed against all lags of all the other columns
    which include numbers. NaN values, introduced by the various shifts, are replaced with 0.

    Parameters
    ----------
    target : str
        Column which is used as target for the cross correlation.

    ingested_data : DataFrame
        DataFrame which contains the various time-series, one for column.

    max_lags : int
        Limit the analysis to max lags.

    modes : [str]
        Cross-correlation can be computed with different algorithms. The available choices are:

        - `matlab_normalized`: same as using the MatLab function xcorr(x, y, 'normalized')
        - `pearson` : use Pearson formula (NaN values are fillled to 0)
        - `kendall`: use Kendall formula (NaN values are filled to 0)
        - `spearman`: use Spearman formula (NaN values are filled to 0)

    Returns
    -------
    result : dict
        Dictionary with a Pandas DataFrame set for every indicated mode.
        Each DataFrame has the lags as index and the correlation value for each column.

    Examples
    --------
    Create some sample time-series.
    >>> dates = pd.date_range('2000-01-01', periods=30)  # Last index is 2000-01-30
    >>> ds = pd.DatetimeIndex(dates, freq="D")
    >>>
    >>> x = np.linspace(0, 2 * np.pi, 60)
    >>> y = np.sin(x)
    >>>
    >>> np.random.seed(0)
    >>> noise = np.random.normal(0, 2.0, 60)
    >>> y = y + noise
    >>>
    >>> a = y[:30]
    >>> b = y[5:35]
    >>>
    >>> timeseries_dataframe = DataFrame(data={"a": a, "b": b}, index=ds)

    Compute the cross-correlation:
    >>> calc_xcorr("a", timeseries_dataframe, 7, ["pearson"])
    {'pearson':  b
             -7  0.316213
             -6 -0.022288
             -5  0.112483
             -4 -0.268724
             -3  0.105511
             -2  0.178658
             -1  0.101505
              0  0.051641
              1 -0.360475
              2 -0.074952
              3 -0.047689
              4 -0.252324
              5  0.796120
              6 -0.170558
              7 -0.009305
    }

    This is expected; the biggest value of cross-correlation is at index `5`. It is true that `b` is exactly time-series
    `a`, but shifted forward of `5` lags.
    """

    def df_shifted(df, _target=None, lag=0):
        if not lag and not _target:
            return df
        new = {}
        for c in df.columns:
            if c == _target:
                new[c] = df[_target]
            else:
                new[c] = df[c].shift(periods=lag)
        return pd.DataFrame(data=new)

    columns = ingested_data.columns.tolist()
    columns = [elem for elem in columns if ingested_data[elem].dtype != str and elem != target]

    results = {}
    for mode in modes:
        result = DataFrame(columns=columns, dtype=np.float64)
        if mode == 'matlab_normalized':
            lags_to_displace = max_lags if max_lags < len(ingested_data) else len(ingested_data) - 1
            for col in columns:
                x = ingested_data[target]
                y = ingested_data[col]

                c = np.correlate(x, y, mode="full")

                # This is needed to obtain the same result of the MatLab `xcorr` function with normalized results.
                # You can find the formula in the function pyplot.xcorr; however, here the property
                # sqrt(x*y) = sqrt(x) * sqrt(y)
                # is applied in order to avoid overflows if the ingested values are particularly high.
                den = np.sqrt(np.dot(x, x)) * np.sqrt(np.dot(y, y))
                c = np.divide(c, den)

                # This assigns the correct indexes to the results.
                if len(ingested_data) > max_lags:
                    c = c[len(ingested_data) - 1 - max_lags:len(ingested_data) + max_lags]

                result[col] = c

            result.index -= lags_to_displace

        elif mode == 'granger':
            for col in columns:
                granger_max_lags = int(len(ingested_data) / 3) - 1
                granger_max_lags = granger_max_lags if granger_max_lags < max_lags else max_lags

                # Trick to compute both negative and positive lags
                df = ingested_data[[col, target]]
                granger_result = grangercausalitytests(df, maxlag=granger_max_lags, verbose=False)
                for i in granger_result:
                    result.loc[-i, col] = 1 - granger_result[i][0]['params_ftest'][1]

                df = ingested_data[[target, col]]
                granger_result = grangercausalitytests(df, maxlag=granger_max_lags, verbose=False)
                for i in granger_result:
                    result.loc[i, col] = 1 - granger_result[i][0]['params_ftest'][1]

            result.sort_index(inplace=True)

        else:
            lags_to_displace = max_lags if max_lags < len(ingested_data) else len(ingested_data)
            for i in range(-lags_to_displace, lags_to_displace + 1):
                shifted = df_shifted(ingested_data, target, i)
                shifted.fillna(0, inplace=True)

                corr = [shifted[target].corr(other=shifted[col], method=mode) for col in columns]
                result.loc[i] = corr

        results[mode] = result

    return results


def calc_all_xcorr(ingested_data: DataFrame, param_config: dict) -> dict:
    """
    Compute, for every column in `ingested_data` (excluding the index) the cross-correlation of that series with respect
    to all others columns in ingested data.

    Parameters
    ----------
    ingested_data : DataFrame
        Pandas DataFrame for which the cross-correlation of all columns should be computed.

    param_config : dict
        TIMEX configuration dictionary, needed to for `xcorr_parameters`.
        In the `xcorr_parameters` sub-dictionary, `xcorr_modes` and `xcorr_max_lags` will be used.
        `xcorr_modes` indicate the different algorithms which should be used to compute the cross-correlation.
        The available choices are:

        - `matlab_normalized`: same as using the MatLab function xcorr(x, y, 'normalized')
        - `pearson` : use Pearson formula (NaN values are fillled to 0)
        - `kendall`: use Kendall formula (NaN values are filled to 0)
        - `spearman`: use Spearman formula (NaN values are filled to 0)

        `xcorr_max_lags` is the number of lags, both in positive and negative direction, to which the cross-correlation
        calculations should be limited to.

    Returns
    -------
    dict
        Python dict with a key for every time-series in `ingested_data`; every key will correspond to another dictionary
        with one entry for each cross-correlation algorithm requested.

    Examples
    --------
    Create sample data.
    >>> dates = pd.date_range('2000-01-01', periods=30)  # Last index is 2000-01-30
    >>> ds = pd.DatetimeIndex(dates, freq="D")
    >>>
    >>> x = np.linspace(0, 2 * np.pi, 60)
    >>> y = np.sin(x)
    >>> np.random.seed(0)
    >>> noise = np.random.normal(0, 2.0, 60)
    >>> y = y + noise
    >>>
    >>> a = y[:30]
    >>> b = y[2:32]
    >>> c = y[4:34]
    >>>
    >>> timeseries_dataframe = DataFrame(data={"a": a, "b": b, "c": c}, index=ds)

    Compute the cross-correlation on this DataFrame:
    >>> param_config = {
    >>>     "xcorr_parameters": {
    >>>         "xcorr_max_lags": 2,
    >>>         "xcorr_mode": "pearson,matlab_normalized"
    >>>     }
    >>> }
    >>> calc_all_xcorr(timeseries_dataframe, param_config)
    {'a': {'pearson':            b         c
                             -2 -0.252086  0.117286
                             -1  0.006370  0.064624
                              0 -0.011866 -0.290049
                              1 -0.115114 -0.091762
                              2  0.951782 -0.024158,
           'matlab_normalized':  b         c
                             -2  0.109634  0.287681
                             -1  0.314318  0.239430
                             0   0.319016  0.008663
                             1   0.244525  0.197663
                             2   0.965095  0.260254},

     'b': {'pearson':            a         c
                             -2  0.998491 -0.353341
                             -1 -0.085531 -0.007476
                              0 -0.011866  0.048841
                              1  0.013242 -0.092448
                              2 -0.258411  0.895226,
           'matlab_normalized':  a         c
                             -2  0.965095 -0.063331
                             -1  0.244525  0.177921
                             0   0.319016  0.252201
                             1   0.314318  0.183260
                             2   0.109634  0.862899},

     'c': {'pearson':            a         b
                             -2  0.076014  0.929572
                             -1 -0.013978 -0.026488
                              0 -0.290049  0.048841
                              1  0.038452 -0.043913
                              2  0.125275 -0.354749,
           'matlab_normalized':  a         b
                             -2  0.260254  0.862899
                             -1  0.197663  0.183260
                             0   0.008663  0.252201
                             1   0.239430  0.177921
                             2   0.287681 -0.063331}}
    """
    xcorr_max_lags = param_config['xcorr_parameters']['xcorr_max_lags']
    xcorr_modes = [*param_config['xcorr_parameters']["xcorr_mode"].split(",")]
    d = {}
    for col in ingested_data.columns:
        d[col] = calc_xcorr(col, ingested_data, max_lags=xcorr_max_lags, modes=xcorr_modes)

    return d


import logging
import gettext
import pathlib
import os

import pandas
from pandas import Grouper, DataFrame
import plotly.graph_objects as go
import numpy as np

from dash import dcc
from dash import html
from plotly.subplots import make_subplots
import networkx as nx
import dash_bootstrap_components as dbc

from colorhash import ColorHash
from statsmodels.tsa.seasonal import seasonal_decompose

from timexseries.data_prediction import ValidationPerformance
from timexseries.data_prediction.models.predictor import SingleResult
import calendar

from timexseries.timeseries_container import TimeSeriesContainer

log = logging.getLogger(__name__)

# Default method to get a translated text.
_ = lambda x: x


def create_timeseries_dash_children(timeseries_container: TimeSeriesContainer, param_config: dict):
    """
    Creates the Dash children for a specific time-series. They include a line plot, histogram, box plot and
    autocorrelation plot. For each model on the time-series the prediction plot and performance plot are also added.

    Cross-correlation plots and graphs are shown, if the the `timeseries_container` have it.

    If the `timeseries_container` also have the `historical_prediction`, it is shown in a plot.
    
    Parameters
    ----------
    timeseries_container: TimeSeriesContainer
        Time-series for which the various plots and graphs will be returned.
    
    param_config : dict
        TIMEX configuration parameters dictionary, used for `visualization_parameters` which contains settings to
        customize some plots and graphs.

    Returns
    -------
    list
        List of Dash children.

    Examples
    --------
    Given a `timexseries.timeseries_container.TimeSeriesContainer` object, obtained for example through
    `timexseries.data_prediction.pipeline.create_timeseries_containers`, create all the Dash object which could be shown in a
    Dash app:
    >>> param_config = {
    ...  "input_parameters": {},
    ...  "model_parameters": {
    ...      "models": "fbprophet",
    ...      "possible_transformations": "none,log_modified",
    ...      "main_accuracy_estimator": "mae",
    ...      "delta_training_percentage": 20,
    ...      "test_values": 5,
    ...      "prediction_lags": 7,
    ...  },
    ...  "historical_prediction_parameters": {
    ...      "initial_index": "2000-01-25",
    ...      "save_path": "example.pkl"
    ...  },
    ...  "visualization_parameters": {}
    ...}
    >>>  plots = create_timeseries_dash_children(timeseries_container, param_config)
    """
    children = []

    visualization_parameters = param_config["visualization_parameters"]
    timeseries_data = timeseries_container.timeseries_data

    name = timeseries_data.columns[0]

    locale_dir = pathlib.Path(os.path.abspath(__file__)).parent / "locales"

    global _
    try:
        gt = gettext.translation('messages', localedir=locale_dir, languages=[visualization_parameters["language"]])
        gt.install()
        _ = gt.gettext
    except:
        gt = gettext.translation('messages', localedir=locale_dir, languages=['en'])
        gt.install()
        _ = gt.gettext

    # Data visualization with plots
    children.extend([
        html.H2(children=name + _(' analysis'), id=name),
        html.H3(_("Data visualization")),
        line_plot(timeseries_data),
        histogram_plot(timeseries_data),
        box_plot(timeseries_data, visualization_parameters),
        box_plot_aggregate(timeseries_data, visualization_parameters),
        components_plot(timeseries_data),
        autocorrelation_plot(timeseries_data),
    ])

    # Plot cross-correlation plot and graphs, if requested.
    if timeseries_container.xcorr is not None:
        graph_corr_threshold = visualization_parameters[
            "xcorr_graph_threshold"] if "xcorr_graph_threshold" in visualization_parameters else None

        children.extend([
            html.H3(_("Cross-correlation")),
            html.Div(_("Negative lags (left part) show the correlation between this scenario and the future of the "
                       "others.")),
            html.Div(_("Meanwhile, positive lags (right part) shows the correlation between this scenario "
                       "and the past of the others.")),
            cross_correlation_plot(timeseries_container.xcorr),
            html.Div(_("The peaks found using each cross-correlation modality are shown in the graphs:")),
            cross_correlation_graph(name, timeseries_container.xcorr, graph_corr_threshold)
        ])

    # Plot the prediction results, if requested.
    if timeseries_container.models is not None:
        model_parameters = param_config["model_parameters"]

        models = timeseries_container.models

        children.append(
            html.H3(_("Training & Validation results")),
        )

        for model_name in models:
            model = models[model_name]
            model_results = model.results
            model_characteristic = model.characteristics

            test_values = model_characteristic["test_values"]
            main_accuracy_estimator = model_parameters["main_accuracy_estimator"]
            model_results.sort(key=lambda x: getattr(x.testing_performances, main_accuracy_estimator.upper()))

            best_prediction = model_results[0].prediction
            testing_performances = [x.testing_performances for x in model_results]

            children.extend([
                html.H4(f"{model_name}"),
                characteristics_list(model_characteristic, testing_performances),
                # html.Div("Testing performance:"),
                # html.Ul([html.Li(key + ": " + str(testing_performances[key])) for key in testing_performances]),
                prediction_plot(timeseries_data, best_prediction, test_values),
                performance_plot(timeseries_data, best_prediction, testing_performances, test_values),
            ])

            # EXTRA
            # Warning: this will plot every model result, with every training set used!
            # children.extend(plot_every_prediction(ingested_data, model_results, main_accuracy_estimator, test_values))

    if timeseries_container.historical_prediction is not None:
        children.extend([
            html.H3(_("Prediction")),
            html.Div(_("For every model the best predictions for each past date are plotted."))
        ])
        for model in timeseries_container.historical_prediction:
            children.extend([
                html.H4(f"{model}"),
                historical_prediction_plot(timeseries_data, timeseries_container.historical_prediction[model],
                                           timeseries_container.models[model].best_prediction)
            ])

    return children


def create_dash_children(timeseries_containers: [TimeSeriesContainer], param_config: dict):
    """
    Create Dash children, in order, for a list of `timexseries.timeseries_container.TimeSeriesContainer`.

    Parameters
    ----------
    timeseries_containers : [TimeSeriesContainer]
        Time-series for which all the plots and graphs will be created.

    param_config : dict
        TIMEX configuration parameters dictionary.

    Returns
    -------
    list
        List of Dash children.

    """
    children = []
    for s in timeseries_containers:
        children.extend(create_timeseries_dash_children(s, param_config))

    return children


def line_plot(df: DataFrame) -> dcc.Graph:
    """
    Create and return the line plot for a dataframe.

    Parameters
    ----------
    df : DataFrame
        Dataframe to plot.

    Returns
    -------
    g : dcc.Graph
        Dash object containing the line plot.

    Examples
    --------
    Get the `figure` attribute if you want to display this in a Jupyter notebook.
    >>> line_plot = line_plot(timeseries_container.timeseries_data).figure
    >>> line_plot.show()
    """
    fig = go.Figure(data=go.Scatter(x=df.index, y=df.iloc[:, 0], mode='lines+markers'))
    fig.update_layout(title=_('Line plot'), xaxis_title=df.index.name, yaxis_title=df.columns[0])

    g = dcc.Graph(
        figure=fig
    )
    return g


def line_plot_multiIndex(df: DataFrame) -> dcc.Graph:
    """
    Returns a line plot for a dataframe with a MultiIndex.
    It is assumed that the first-level index is the real index, and that data should be grouped using the
    second-level one.

    Parameters
    ----------
    df : DataFrame
        Dataframe to plot. It is a multiIndex dataframe.

    Returns
    -------
    g : dcc.Graph
    """
    fig = go.Figure()
    for region in df.index.get_level_values(1).unique():
        fig.add_trace(go.Scatter(x=df.index.get_level_values(0).unique(), y=df.loc[
            (df.index.get_level_values(1) == region), df.columns[0]], name=region))

    fig.update_layout(title=_('Line plot'), xaxis_title=df.index.get_level_values(0).name,
                      yaxis_title=df.columns[0])
    g = dcc.Graph(
        figure=fig
    )
    return g


def histogram_plot(df: DataFrame) -> dcc.Graph:
    """
    Create and return the histogram plot for a dataframe.

    Parameters
    ----------
    df : DataFrame
        Dataframe to plot.

    Returns
    -------
    g : dcc.Graph

    Examples
    --------
    Get the `figure` attribute if you want to display this in a Jupyter notebook.
    >>> hist_plot = hist_plot(timeseries_container.timeseries_data).figure
    >>> hist_plot.show()
    """

    max_value = abs(df.iloc[:, 0].max() - df.iloc[:, 0].min())
    step = max_value / 400

    fig = go.Figure(data=[go.Histogram(x=df.iloc[:, 0])])
    fig.layout.sliders = [dict(
        steps=[dict(method='restyle', args=['xbins.size', i]) for i in np.arange(step, max_value / 2, step)],
        font=dict(color="rgba(0,0,0,0)"),
        tickcolor="rgba(0,0,0,0)"
    )]

    fig.update_layout(title=_('Histogram'), xaxis_title_text=df.columns[0], yaxis_title_text=_('Count'))
    g = dcc.Graph(
        figure=fig
    )
    return g


def components_plot(ingested_data: DataFrame) -> html.Div:
    """
    Create and return the plots of all the components of the time series: level, trend, residual.
    It uses both an additive and multiplicative model, with a subplot.

    Parameters
    ----------
    ingested_data : DataFrame
        Original time series values.

    Returns
    -------
    g : dcc.Graph

    Examples
    --------
    Get the `figure` attribute if you want to display this in a Jupyter notebook.
    >>> comp_plot = components_plot(timeseries_container.timeseries_data)[0].figure
    >>> comp_plot.show()
    """
    modes = ["additive", "multiplicative"]

    fig = make_subplots(
        rows=3,
        cols=1,
        subplot_titles=[_("Trend"), _("Seasonality"), _("Residual")], shared_xaxes=True, vertical_spacing=0.05,
        specs=[[{"secondary_y": True}], [{"secondary_y": True}], [{"secondary_y": True}]]
    )

    interpolated = ingested_data.interpolate()
    interpolated = interpolated.fillna(0)

    for mode in modes:
        try:
            result = seasonal_decompose(interpolated, model=mode)
            trend = result.trend
            seasonal = result.seasonal
            residual = result.resid

            secondary_y = False if mode == "additive" else True

            fig.add_trace(go.Scatter(x=trend.index, y=trend,
                                     mode='lines+markers',
                                     name=_(mode.capitalize()), legendgroup=_(mode.capitalize()),
                                     line=dict(color=ColorHash(mode).hex)),
                          row=1, col=1, secondary_y=secondary_y)
            fig.add_trace(go.Scatter(x=seasonal.index, y=seasonal,
                                     mode='lines+markers', showlegend=False,
                                     name=_(mode.capitalize()), legendgroup=_(mode.capitalize()),
                                     line=dict(color=ColorHash(mode).hex)),
                          row=2, col=1, secondary_y=secondary_y)
            fig.add_trace(go.Scatter(x=residual.index, y=residual,
                                     mode='lines+markers', showlegend=False,
                                     name=_(mode.capitalize()), legendgroup=_(mode.capitalize()),
                                     line=dict(color=ColorHash(mode).hex)),
                          row=3, col=1, secondary_y=secondary_y)
        except ValueError:
            log.warning(f"Multiplicative decomposition not available for {ingested_data.columns[0]}")

    fig.update_layout(title=_("Components decomposition"), height=1000, legend_title_text=_('Decomposition model'))
    fig.update_yaxes(title_text="<b>" + _('Additive') + "</b>", secondary_y=False)
    fig.update_yaxes(title_text="<b>" + _('Multiplicative') + "</b>", secondary_y=True)

    g = dcc.Graph(
        figure=fig
    )

    warning = html.H5(_("Multiplicative model is not available for series which contain zero or negative values."))

    return html.Div([g, warning])


def autocorrelation_plot(df: DataFrame) -> dcc.Graph:
    """
    Create and return the autocorrelation plot for a dataframe.

    Parameters
    ----------
    df : DataFrame
        Dataframe to use in the autocorrelation plot.

    Returns
    -------
    g : dcc.Graph

    Examples
    --------
    Get the `figure` attribute if you want to display this in a Jupyter notebook.
    >>> auto_plot = autocorrelation_plot(timeseries_container.timeseries_data).figure
    >>> auto_plot.show()
    """
    # Code from https://github.com/pandas-dev/pandas/blob/v1.1.4/pandas/plotting/_matplotlib/misc.py
    n = len(df)
    data = np.asarray(df)
    mean = np.mean(data)
    c0 = np.sum((data - mean) ** 2) / float(n)

    def r(h):
        return ((data[: n - h] - mean) * (data[h:] - mean)).sum() / float(n) / c0

    x = np.arange(n) + 1
    y = [r(loc) for loc in x]

    z95 = 1.959963984540054
    z99 = 2.5758293035489004

    c1 = z99 / np.sqrt(n)
    c2 = z95 / np.sqrt(n)
    c3 = -z95 / np.sqrt(n)
    c4 = -z99 / np.sqrt(n)

    fig = go.Figure()
    fig.add_trace(go.Scatter(x=x, y=y, mode='lines', name=_('autocorrelation')))
    fig.add_trace(go.Scatter(x=x, y=np.full(n, c1), line=dict(color='gray', width=1), name='z99'))
    fig.add_trace(go.Scatter(x=x, y=np.full(n, c2), line=dict(color='gray', width=1), name='z95'))
    fig.add_trace(go.Scatter(x=x, y=np.full(n, c3), line=dict(color='gray', width=1), name='-z95'))
    fig.add_trace(go.Scatter(x=x, y=np.full(n, c4), line=dict(color='gray', width=1), name='-z99'))
    fig.update_layout(title=_('Autocorrelation plot'), xaxis_title=_('Lags'), yaxis_title=_('Autocorrelation'))
    fig.update_yaxes(tick0=-1.0, dtick=0.25)
    fig.update_yaxes(range=[-1.2, 1.2])
    g = dcc.Graph(
        figure=fig
    )
    return g


def cross_correlation_plot(xcorr: dict):
    """
    Create and return the cross-correlation plot for all the columns in the dataframe.
    The time-series column is used as target; the correlation is shown in a subplot for every modality used to compute
    the x-correlation.

    Parameters
    ----------
    xcorr : dict
        Cross-correlation values.

    Returns
    -------
    g : dcc.Graph

    Examples
    --------
    Get the `figure` attribute if you want to display this in a Jupyter notebook.
    >>> xcorr_plot = cross_correlation_plot(timeseries_container.xcorr).figure
    >>> xcorr_plot.show()
    """
    subplots = len(xcorr)
    combs = [(1, 1), (1, 2), (2, 1), (2, 2)]

    rows = 1 if subplots < 3 else 2
    cols = 1 if subplots < 2 else 2

    fig = make_subplots(
        rows=rows, cols=cols,
        subplot_titles=([*xcorr.keys()]))

    i = 0
    for mode in xcorr:
        for col in xcorr[mode].columns:
            fig.add_trace(go.Scatter(x=xcorr[mode].index, y=xcorr[mode][col],
                                     mode='lines',
                                     name=col, legendgroup=col, line=dict(color=ColorHash(col).hex),
                                     showlegend=True if i == 0 else False),
                          row=combs[i][0], col=combs[i][1])
        i += 1

    # Formula from https://support.minitab.com/en-us/minitab/18/help-and-how-to/modeling-statistics/time-series/how-to/cross-correlation/interpret-the-results/all-statistics-and-graphs/
    # significance_level = DataFrame(columns=['Value'], dtype=np.float64)
    # for i in range(-lags, lags):
    #     significance_level.loc[i] = 2 / np.sqrt(lags - abs(i))

    # fig.add_trace(
    #     go.Scatter(x=significance_level.index, y=significance_level['Value'], line=dict(color='gray', width=1), name='z95'))
    # fig.add_trace(
    #     go.Scatter(x=significance_level.index, y=-significance_level['Value'], line=dict(color='gray', width=1), name='-z95'))

    fig.update_layout(title=_("Cross-correlation using different algorithms"))
    fig.update_xaxes(title_text=_("Lags"))
    fig.update_yaxes(tick0=-1.0, dtick=0.25, range=[-1.2, 1.2], title_text=_("Correlation"))

    g = dcc.Graph(
        figure=fig
    )
    return g


def cross_correlation_graph(name: str, xcorr: dict, threshold: float = 0) -> dcc.Graph:
    """
    Create and return the cross-correlation graphs for all the columns in the dataframe.
    A graph is created for each mode used to compute the x-correlation.

    The nodes are all the time-series which can be found in `xcorr`; an arc is drawn from `target` node to another node
    if the cross-correlation with that time-series, at any lag, is above the `threshold`. The arc contains also the
    information on the lag.

    Parameters
    ----------
    name : str
        Name of the target.

    xcorr : dict
        Cross-correlation dataframe.

    threshold : int
        Minimum value of correlation for which a edge should be drawn. Default 0.

    Returns
    -------
    g : dcc.Graph

    Examples
    --------
    This is thought to be shown in a Dash app, so it could be difficult to show in Jupyter.
    >>> xcorr_graph = cross_correlation_graph("a", timeseries_container.xcorr, 0.7)
    """
    figures = []

    i = 0
    for mode in xcorr:
        G = nx.DiGraph()
        G.add_nodes_from(xcorr[mode].columns)
        G.add_node(name)

        for col in xcorr[mode].columns:
            index_of_max = xcorr[mode][col].abs().idxmax()
            corr = xcorr[mode].loc[index_of_max, col]
            if abs(corr) > threshold:
                G.add_edge(name, col, corr=corr, lag=index_of_max)

        pos = nx.layout.spring_layout(G)

        # Create Edges
        edge_trace = go.Scatter(
            x=[],
            y=[],
            line=dict(color='black'),
            mode='lines',
            hoverinfo='skip',
        )

        for edge in G.edges():
            start = edge[0]
            end = edge[1]
            x0, y0 = pos.get(start)
            x1, y1 = pos.get(end)
            edge_trace['x'] += tuple([x0, x1, None])
            edge_trace['y'] += tuple([y0, y1, None])

        # Create Nodes
        node_trace = go.Scatter(
            x=[],
            y=[],
            mode='markers+text',
            text=[node for node in G.nodes],
            textposition="bottom center",
            hoverinfo='skip',
            marker=dict(
                color='green',
                size=15)
        )

        for node in G.nodes():
            x, y = pos.get(node)
            node_trace['x'] += tuple([x])
            node_trace['y'] += tuple([y])

        # Annotations to support arrows
        edges_positions = [e for e in G.edges]
        annotateArrows = [dict(showarrow=True, arrowsize=1.0, arrowwidth=2, arrowhead=2, standoff=2, startstandoff=2,
                               ax=pos[arrow[0]][0], ay=pos[arrow[0]][1], axref='x', ayref='y',
                               x=pos[arrow[1]][0], y=pos[arrow[1]][1], xref='x', yref='y',
                               text="bla") for arrow in edges_positions]

        graph = go.Figure(data=[node_trace, edge_trace],
                          layout=go.Layout(title=str(mode),
                                           xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                                           yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                                           showlegend=False,
                                           annotations=annotateArrows,
                                           height=400, margin=dict(l=10, r=10, t=50, b=30)))

        # Add annotations on edges
        for e in G.edges:
            lag = str(G.edges[e]['lag'])
            corr = str(round(G.edges[e]['corr'], 3))

            end = e[1]
            x, y = pos.get(end)

            graph.add_annotation(x=x, y=y, text=_("Lag: ") + lag + ", corr: " + corr, yshift=20, showarrow=False,
                                 bgcolor='white')

        figures.append(graph)
        i += 1

    n_graphs = len(figures)
    if n_graphs == 1:
        g = dcc.Graph(figure=figures[0])
    elif n_graphs == 2:
        g = html.Div(dbc.Row([
            dbc.Col(dcc.Graph(figure=figures[0])),
            dbc.Col(dcc.Graph(figure=figures[1]))
        ]))
    elif n_graphs == 3:
        g = html.Div([
            dbc.Row([
                dbc.Col(dcc.Graph(figure=figures[0])),
                dbc.Col(dcc.Graph(figure=figures[1]))
            ]),
            dbc.Row([
                dbc.Col(dcc.Graph(figure=figures[2]))
            ])
        ])
    elif n_graphs == 4:
        g = html.Div([
            dbc.Row([
                dbc.Col(dcc.Graph(figure=figures[0])),
                dbc.Col(dcc.Graph(figure=figures[1])),
            ]),
            dbc.Row([
                dbc.Col(dcc.Graph(figure=figures[2])),
                dbc.Col(dcc.Graph(figure=figures[3]))
            ])
        ])
    else:
        g = html.Div()

    return g


def box_plot(df: DataFrame, visualization_parameters: dict) -> dcc.Graph:
    """
    Create and return the box-and-whisker plot for a dataframe.

    Parameters
    ----------
    df : DataFrame
        Dataframe to use in the box plot.

    visualization_parameters : dict
        Options set by the user. In particular, `box_plot_frequency` is used to determine the number of boxes. The
        default value is `1W`.

    Returns
    -------
    g : dcc.Graph

    Examples
    --------
    Get the `figure` attribute if you want to display this in a Jupyter notebook.
    >>> box_plot = box_plot(timeseries_container.timeseries_data, param_config["visualization_parameters"]).figure
    >>> box_plot.show()
    """
    temp = df.iloc[:, 0]

    aggregations = ["1W", "2W", "1M", "3M", "4M", "6M", "1Y"]

    try:
        initial_freq = visualization_parameters['box_plot_frequency']
    except KeyError:
        initial_freq = '1W'

    traces = []

    for freq in aggregations:
        is_visible = True if initial_freq == freq else 'legendonly'

        # Small trick needed to show `name` in legend.
        traces.append(go.Scatter(x=[None], y=[None], legendgroup=freq, name=freq,
                                 visible=is_visible))

    for freq in aggregations:
        groups = temp.groupby(Grouper(freq=freq))
        is_visible = True if initial_freq == freq else 'legendonly'

        for group in groups:
            traces.append(go.Box(
                name=str(group[0].normalize()),
                y=group[1], legendgroup=freq, showlegend=False, visible=is_visible, xperiodalignment='end'
            ))

    fig = go.Figure(traces)
    fig.update_layout(
        updatemenus=[dict(
            type='buttons',
            direction='right',
            yanchor='bottom',
            xanchor='left',
            y=1,
            active=aggregations.index(_(initial_freq)),
            buttons=list(
                [dict(label=f, method='update', args=[{'visible': [tr.legendgroup == f for tr in traces]}])
                 for f in aggregations]))])

    fig.update_layout(title=_('Box plot'), xaxis_title=df.index.name, yaxis_title=_('Count'), showlegend=False)
    fig.update_xaxes(type='category')
    fig.add_annotation(text=_("The index refer to the end of the period."),
                       xref="paper", yref="paper",
                       x=1.0, y=1.0, yanchor='bottom', showarrow=False)

    g = dcc.Graph(
        figure=fig
    )
    return g


def box_plot_aggregate(df: DataFrame, visualization_parameters: dict) -> dcc.Graph:
    """
    Create and return the aggregate box plot for a dataframe, i.e. a box plot which shows, for each day of the week/for
    each month of the year the distribution of the values. Now also with aggregation over minute, hour, week, month,
    year.

    Parameters
    ----------
    df : DataFrame
        Dataframe to use in the box plot.

    visualization_parameters : dict
        Options set by the user. In particular, `aggregate_box_plot_frequency` is used. Default `weekday`. It controls
        which frequency is activated at launch.

    Returns
    -------
    g : dcc.Graph

    Examples
    --------
    Get the `figure` attribute if you want to display this in a Jupyter notebook.
    >>> abox_plot = box_plot_aggregate(timeseries_container.timeseries_data,
    ...                                param_config["visualization_parameters"]).figure
    >>> abox_plot.show()
    """
    aggregations = ["minute", "hour", "weekday", "day", "month", "year"]
    translated_aggregations = [_("minute"), _("hour"), _("weekday"), _("day"), _("month"), _("year")]

    temp = df.iloc[:, 0]

    try:
        initial_freq = visualization_parameters['aggregate_box_plot_frequency']
    except KeyError:
        initial_freq = 'weekday'

    traces = []

    for freq, translated_freq in zip(aggregations, translated_aggregations):
        is_visible = True if initial_freq == freq else False

        # Small trick needed to show `name` in legend.
        traces.append(go.Scatter(x=[None], y=[None], legendgroup=translated_freq, name=translated_freq,
                                 visible=is_visible))

    for freq, translated_freq in zip(aggregations, translated_aggregations):
        groups = temp.groupby(getattr(temp.index, freq))
        is_visible = True if initial_freq == freq else False

        if freq == "weekday":
            for group in groups:
                traces.append(go.Box(
                    name=calendar.day_name[group[0]],
                    y=group[1], legendgroup=translated_freq, showlegend=False, visible=is_visible
                ))
        elif freq == "month":
            for group in groups:
                traces.append(go.Box(
                    name=calendar.month_name[group[0]],
                    y=group[1], legendgroup=translated_freq, showlegend=False, visible=is_visible
                ))
        else:
            for group in groups:
                traces.append(go.Box(
                    name=group[0],
                    y=group[1], legendgroup=translated_freq, showlegend=False, visible=is_visible
                ))

    fig = go.Figure(data=traces)

    fig.update_layout(
        updatemenus=[dict(
            type='buttons',
            direction='right',
            yanchor='bottom',
            xanchor='left',
            y=1,
            active=translated_aggregations.index(_(initial_freq)),
            buttons=list(
                [dict(label=f, method='update', args=[{'visible': [tr.legendgroup == f for tr in traces]}])
                 for f in translated_aggregations]))])

    fig.update_layout(title=_('Aggregate box plot'), yaxis_title=_('Count'), showlegend=False)
    fig.update_xaxes(type='category')

    g = dcc.Graph(
        figure=fig
    )
    return g


def prediction_plot(df: DataFrame, predicted_data: DataFrame, test_values: int = 0) -> dcc.Graph:
    """
    Create and return a plot which contains the prediction for a dataframe.
    The plot is built using two dataframe: `ingested_data` and `predicted_data`.

    `ingested_data` includes the raw data ingested by the app, while `predicted_data` contains the actual prediction
    made by a model.

    Note that `predicted_data` starts at the first value used for training.

    The data not used for training is plotted in black, the data used for training is plotted in green and the
    validation values dashed.

    Note that `predicted_data` may or not have the columns "yhat_lower" and "yhat_upper".

    Parameters
    ----------
    df : DataFrame
        Raw values ingested by the app.

    predicted_data : DataFrame
        Prediction created by a model.

    test_values : int, optional, default 0
        Number of validation values used in the testing.

    Returns
    -------
    g : dcc.Graph

    See Also
    --------
    Check `create_timeseries_dash_children` to check the use.
    """
    fig = go.Figure()

    not_training_data = df.loc[:predicted_data.index[0]]
    training_data = df.loc[predicted_data.index[0]:]
    training_data = training_data.iloc[:-test_values]

    fig.add_trace(go.Scatter(x=not_training_data.index, y=not_training_data.iloc[:, 0],
                             line=dict(color='black'),
                             mode='markers',
                             name=_('unused data')))
    fig.add_trace(go.Scatter(x=training_data.index, y=training_data.iloc[:, 0],
                             line=dict(color='green', width=4, dash='dash'),
                             mode='markers',
                             name=_('training data'),
                             ))

    if test_values > 0:
        test_data = df.iloc[-test_values:]
        fig.add_trace(go.Scatter(x=test_data.index, y=test_data.iloc[:, 0],
                                 line=dict(color='green', width=3, dash='dot'),
                                 name=_('validation data')))

    fig.add_trace(go.Scatter(x=predicted_data.index, y=predicted_data['yhat'],
                             line=dict(color='blue'),
                             mode='lines+markers',
                             name=_('yhat')))
    try:
        fig.add_trace(go.Scatter(x=predicted_data.index, y=predicted_data['yhat_lower'],
                                 line=dict(color='lightgreen', dash='dash'),
                                 name=_('yhatlower')))
        fig.add_trace(go.Scatter(x=predicted_data.index, y=predicted_data['yhat_upper'],
                                 line=dict(color='lightgreen', dash='dash'),
                                 name=_('yhatupper')))
    except:
        pass

    fig.update_layout(title=_("Best prediction for the validation set"), xaxis_title=df.index.name,
                      yaxis_title=df.columns[0])
    g = dcc.Graph(
        figure=fig
    )
    return g


def historical_prediction_plot(real_data: DataFrame, historical_prediction: DataFrame,
                               future_prediction: DataFrame) -> html.Div:
    """
    Create and return a plot which contains the best prediction found by this model for this time series, along with
    the historical prediction. The plot of the error is also drawn.

    Note that `predicted_data` may or not have the columns "yhat_lower" and "yhat_upper".

    Parameters
    ----------
    real_data : DataFrame
        Raw values ingested by the app.

    historical_prediction : DataFrame
        Historical prediction.

    future_prediction : DataFrame
        Best prediction, corresponding to the `best_prediction` attribute of a
        `timexseries.data_prediction.models.predictor.ModelResult`.

    Returns
    -------
    g : dcc.Graph

    See Also
    --------
    Check `create_timeseries_dash_children` to check the use.
    """
    new_children = []
    fig = make_subplots(rows=2, cols=1, shared_xaxes=True, vertical_spacing=0.02)

    timeseries_name = real_data.columns[0]
    first_predicted_index = historical_prediction.index[0]
    last_real_index = real_data.index[-1]

    validation_real_data = real_data.loc[first_predicted_index:, timeseries_name]

    testing_performance = ValidationPerformance(first_predicted_index)
    testing_performance.set_testing_stats(actual=validation_real_data,
                                          predicted=historical_prediction.loc[:last_real_index, timeseries_name])
    new_children.extend([
        html.Div(_("This model, during the history, reached these performances on unseen data:")),
        show_errors(testing_performance),
        html.Div(_("Range of validation data: ") + str(validation_real_data.max() - validation_real_data.min()))])

    fig.add_trace(go.Scatter(x=real_data.index, y=real_data.iloc[:, 0],
                             line=dict(color='red'),
                             mode='markers',
                             name=_('real data')), row=1, col=1)

    fig.add_trace(go.Scatter(x=historical_prediction.index, y=historical_prediction.iloc[:, 0],
                             line=dict(color='blue'),
                             mode='lines+markers',
                             name=_('historical prediction')), row=1, col=1)

    future_prediction.loc[historical_prediction.index[-1], 'yhat'] = historical_prediction.iloc[-1, 0]
    future_prediction = future_prediction.loc[historical_prediction.index[-1]:, :]

    fig.add_trace(go.Scatter(x=future_prediction.index, y=future_prediction['yhat'],
                             line=dict(color='lightgreen'),
                             mode='lines+markers',
                             name=_('future yhat')), row=1, col=1)

    error_series = validation_real_data - historical_prediction.iloc[:, 0]
    fig.add_trace(go.Scatter(x=error_series.index, y=error_series,
                             line=dict(color='black'),
                             mode='lines+markers',
                             name=_("Error series")), row=2, col=1)

    fig.update_yaxes(title_text=_("Historical prediction"), row=1, col=1)
    fig.update_yaxes(title_text=_("Error series"), row=2, col=1)

    fig.update_xaxes(title_text=real_data.index.name, row=2, col=1)
    fig.update_layout(title=_("Historical prediction"), height=900)

    g = dcc.Graph(
        figure=fig
    )
    new_children.append(g)

    fig_hist = go.Figure()
    fig_hist.add_trace(go.Histogram(x=error_series, histnorm="probability",
                                    marker=dict(color='black'), name=_("Error series histogram")))
    error_max_value = abs(error_series.max() - error_series.min())
    step = error_max_value / 200
    fig_hist.layout.sliders = [dict(
        steps=[dict(method='restyle', args=['xbins.size', i]) for i in np.arange(step, error_max_value / 2, step)],
        font=dict(color="rgba(0,0,0,0)"),
        tickcolor="rgba(0,0,0,0)"
    )]
    fig_hist.update_yaxes(title_text=_("Error series histogram"), tickformat='.1%')
    fig_hist.update_xaxes(title_text=timeseries_name)

    g = dcc.Graph(
        figure=fig_hist
    )
    new_children.append(g)

    return html.Div(new_children)


def performance_plot(df: DataFrame, predicted_data: DataFrame, testing_performances: [ValidationPerformance],
                     test_values: int) -> dcc.Graph:
    """
    Create and return the performance plot of the model; for every error kind (i.e. MSE, MAE, etc) plot the values it
    assumes using different training windows.
    Plot the training data in the end.

    Parameters
    ----------
    df : DataFrame
        Raw values ingested by the app.

    predicted_data : DataFrame
        Prediction created by a model.

    testing_performances : [ValidationPerformance]
        List of ValidationPerformance object. Every object is related to a specific training windows, hence
        it shows the performance using that window.

    test_values : int
        Number of values used for testing performance.

    Returns
    -------
    g : dcc.Graph

    See Also
    --------
    Check `create_timeseries_dash_children` to check the use.
    """
    fig = make_subplots(rows=4, cols=1, shared_xaxes=True, vertical_spacing=0.02)

    training_data = df.iloc[:-test_values]

    data_performances = []

    for tp in testing_performances:
        data_performances.append([tp.first_used_index, tp.MAE, tp.MSE, tp.AM])

    df_performances = pandas.DataFrame(data_performances, columns=['index', 'mae', 'mse', 'am'])
    df_performances.set_index('index', drop=True, inplace=True)
    df_performances.sort_index(inplace=True)

    fig.append_trace(go.Scatter(x=df_performances.index, y=df_performances['mae'],
                                line=dict(color='red'),
                                mode="lines+markers",
                                name='MAE'), row=1, col=1)

    fig.append_trace(go.Scatter(x=df_performances.index, y=df_performances['mse'],
                                line=dict(color='green'),
                                mode="lines+markers",
                                name='MSE'), row=2, col=1)

    fig.append_trace(go.Scatter(x=df_performances.index, y=df_performances['am'],
                                line=dict(color='blue'),
                                mode="lines+markers",
                                name='AM'), row=3, col=1)

    fig.append_trace(go.Scatter(x=training_data.index, y=training_data.iloc[:, 0],
                                line=dict(color='black'),
                                mode='markers',
                                name=_('training data')), row=4, col=1)

    # Small trick to make the x-axis have the same length of the "Prediction plot"
    predicted_data.iloc[:, 0] = "nan"
    fig.append_trace(go.Scatter(x=predicted_data.index, y=predicted_data.iloc[:, 0],
                                mode='lines+markers',
                                name='yhat', showlegend=False), row=4, col=1)

    fig.update_yaxes(title_text="MAE", row=1, col=1)
    fig.update_yaxes(title_text="MSE", row=2, col=1)
    fig.update_yaxes(title_text="AM", row=3, col=1)
    fig.update_yaxes(title_text=df.columns[0], row=4, col=1)

    fig.update_layout(title=_('Performances with different training windows'), height=900)
    g = dcc.Graph(
        figure=fig
    )
    return g


def plot_every_prediction(df: DataFrame, model_results: [SingleResult],
                          main_accuracy_estimator: str, test_values: int):
    new_childrens = [html.Div("EXTRA: plot _EVERY_ prediction\n")]

    model_results.sort(key=lambda x: len(x.prediction))

    for r in model_results:
        predicted_data = r.prediction
        testing_performance = r.testing_performances
        plot = prediction_plot(df, predicted_data, test_values)
        plot.figure.update_layout(title="")
        new_childrens.extend([
            html.Div(main_accuracy_estimator.upper()
                     + ": " + str(getattr(testing_performance, main_accuracy_estimator.upper()))),
            plot
        ])

    return new_childrens


def characteristics_list(model_characteristics: dict, testing_performances: [ValidationPerformance]) -> html.Div:
    """
    Create and return an HTML Div which contains a list of natural language characteristic
    relative to a prediction model.

    Parameters
    ----------
    model_characteristics : dict
        key-value for each characteristic to write in natural language.

    testing_performances : [ValidationPerformance]
        Useful to write also information about the testing performances.

    Returns
    -------
    html.Div()
    """

    def get_text_char(key: str, value: any) -> str:
        value = str(value)
        switcher = {
            "name": _("Model type: ") + value,
            "test_values": _('Values used for testing: last ') + value + _(' values'),
            "delta_training_percentage": _('The length of the training windows is the ') + value
                                         + "%" + _(' of the length of the time series.'),
            "delta_training_values": _('Training windows are composed of ') + value + _(' values.'),
            "extra_regressors": _("The model has used ") + value + _(" as extra-regressor(s) to improve the training."),
            "transformation": _('The model has used a ') + value + _(
                ' transformation on the input data.') if value != "none "
            else _('The model has not used any pre/post transformation on input data.')
        }
        return switcher.get(key, "Invalid choice!")

    elems = [html.Div(_('Model characteristics:')),
             html.Ul([html.Li(get_text_char(key, model_characteristics[key])) for key in model_characteristics]),
             html.Div(_("This model, using the best training window, reaches these performances:")),
             show_errors(testing_performances[0])]

    return html.Div(elems)


def show_errors(testing_performances: ValidationPerformance) -> html.Ul:
    """
    Create an HTML list with each error-metric in `testing_performances`.

    Parameters
    ----------
    testing_performances : ValidationPerformance
        Error metrics to show.

    Returns
    -------
    html.Ul
        HTML list with all the error-metrics.
    """
    import math

    def round_n(n: float):
        dec_part, int_part = math.modf(n)

        if abs(int_part) > 1:
            return str(round(n, 3))
        else:
            return format(n, '.3g')

    def get_text_perf(key: str, value: any) -> str:
        switcher = {
            "MAE": "MAE: " + round_n(value),
            "RMSE": "RMSE: " + round_n(value),
            "MSE": "MSE: " + round_n(value),
            "AM": _('Arithmetic mean of errors:') + round_n(value),
            "SD": _('Standard deviation of errors: ') + round_n(value)
        }
        return switcher.get(key, "Invalid choice!")

    testing_performances = testing_performances.get_dict()
    del testing_performances["first_used_index"]

    return html.Ul([html.Li(get_text_perf(key, testing_performances[key])) for key in testing_performances])



import logging

from pandas import DataFrame

from timexseries.data_prediction import PredictionModel
log = logging.getLogger(__name__)


class MockUpModel(PredictionModel):
    """
    Mock up prediction model. Useful for testing purposes.
    This model "simulates" a real model by returning a - at least in dimensions - correct dataframe.
    This dataframe predicts always 0 is no extra regressors have been given, 1 otherwise.

    This can be useful in tests because it runs in very low time and is useful to understand if higher-level functions
    work as intended.
    """

    def __init__(self, params: dict, transformation: str = "none"):
        super().__init__(params, name="MockUp", transformation=transformation)
        try:
            if params["model_parameters"]["mockup_confidence"]:
                self.confidence_intervals = True
            else:
                self.confidence_intervals = False
        except KeyError:
            self.confidence_intervals = False

        try:
            self.forced_predictions = params["model_parameters"]["mockup_forced_predictions"]
        except KeyError:
            self.forced_predictions = None

        # try:
        #     self.forecast_value = params["model_parameters"]["mockup_forecast_value"]
        # except KeyError:
        #     self.forecast_value = 0

        self.extra_regressors_in_training = None
        self.extra_regressors_in_predict = None
        self.len_train_set = 0
        self.requested_predictions = 0

    def train(self, input_data: DataFrame, extra_regressors: DataFrame = None):
        """Overrides PredictionModel.train()"""
        self.extra_regressors_in_training = extra_regressors
        self.len_train_set = len(input_data)

    def predict(self, future_dataframe: DataFrame, extra_regressors: DataFrame = None) -> DataFrame:
        """Overrides PredictionModel.predict()"""
        self.extra_regressors_in_training = extra_regressors
        self.requested_predictions = len(future_dataframe) - self.len_train_set
        self.extra_regressors_in_predict = extra_regressors

        if self.forced_predictions is not None:
            initial_index = future_dataframe.index[0]
            final_index = future_dataframe.index[-1]
            future_dataframe.loc[:, 'yhat'] = self.forced_predictions.loc[initial_index:final_index]
            if self.confidence_intervals:
                future_dataframe.loc[:, 'yhat_lower'] = self.forced_predictions.loc[initial_index:final_index].apply(
                    lambda x: x - 0.5
                )
                future_dataframe.loc[:, 'yhat_upper'] = self.forced_predictions.loc[initial_index:final_index].apply(
                    lambda x: x + 0.5
                )
        else:
            if extra_regressors is None:
                v = 0
            else:
                v = len(extra_regressors.columns)

            future_dataframe.loc[future_dataframe.index[-self.requested_predictions]:, 'yhat'] = v
            if self.confidence_intervals:
                future_dataframe.loc[future_dataframe.index[-self.requested_predictions]:, 'yhat_lower'] = v - 0.5
                future_dataframe.loc[future_dataframe.index[-self.requested_predictions]:, 'yhat_upper'] = v + 0.5

        return future_dataframe.copy()





import logging
import os

from prophet import Prophet
import pandas as pd
from pandas import DataFrame

from timexseries.data_prediction import PredictionModel

logging.getLogger('fbprophet').setLevel(logging.WARNING)
log = logging.getLogger(__name__)


class FBProphetModel(PredictionModel):
    """Facebook's Prophet prediction model."""

    def __init__(self, params: dict, transformation: str = "none"):
        super().__init__(params, name="FBProphet", transformation=transformation)

        # Stuff needed to make Prophet shut up during training.
        self.suppress_stdout_stderr = suppress_stdout_stderr
        self.fbmodel = Prophet()
        try:
            self.fbprophet_parameters = params["model_parameters"]["fbprophet_parameters"]
        except KeyError:
            self.fbprophet_parameters = None

    def train(self, input_data: DataFrame, extra_regressors: DataFrame = None):
        """Overrides PredictionModel.train()"""

        if self.fbprophet_parameters is not None:
            try:
                timeseries_name = input_data.columns[0]
                date_format = self.fbprophet_parameters["holidays_dataframes"]["date_format"]
                holidays = pd.read_csv(self.fbprophet_parameters["holidays_dataframes"][timeseries_name])
                holidays.loc[:, "ds"].apply(lambda x: pd.to_datetime(x, format=date_format))
                self.fbmodel = Prophet(holidays=holidays)
                log.debug(f"Using a dataframe for holidays...")
            except KeyError:
                self.fbmodel = Prophet()

            try:
                holiday_country = self.fbprophet_parameters["holiday_country"]
                self.fbmodel.add_country_holidays(country_name=holiday_country)
                log.debug(f"Set {holiday_country} as country for holiday calendar...")
            except KeyError:
                pass

        else:
            self.fbmodel = Prophet()

        if extra_regressors is not None:
            # We could apply self.transformation also on the extra regressors.
            # From tests, it looks like it doesn't change much/it worsens the forecasts.
            input_data = input_data.join(extra_regressors)
            input_data.reset_index(inplace=True)
            column_indices = [0, 1]
            new_names = ['ds', 'y']
            old_names = input_data.columns[column_indices]
            input_data.rename(columns=dict(zip(old_names, new_names)), inplace=True)
            [self.fbmodel.add_regressor(col) for col in extra_regressors.columns]

        else:
            input_data.reset_index(inplace=True)
            input_data.columns = ['ds', 'y']

        with self.suppress_stdout_stderr():
            self.fbmodel.fit(input_data)

        #######################
        # param_grid = {
        #     'changepoint_prior_scale': [0.001, 0.01, 0.1, 0.5],
        #     'seasonality_prior_scale': [0.01, 0.1, 1.0, 10.0],
        # }
        # param_grid = {
        #     'changepoint_prior_scale': [0.001, 0.01],
        #     'seasonality_prior_scale': [0.01, 0.1],
        # }
        #
        # if extra_regressors is not None:
        #     input_data = input_data.join(extra_regressors)
        #     input_data.reset_index(inplace=True)
        #     column_indices = [0, 1]
        #     new_names = ['ds', 'y']
        #     old_names = input_data.columns[column_indices]
        #     input_data.rename(columns=dict(zip(old_names, new_names)), inplace=True)
        #
        # else:
        #     input_data.reset_index(inplace=True)
        #     input_data.columns = ['ds', 'y']
        #
        # # Generate all combinations of parameters
        # all_params = [dict(zip(param_grid.keys(), v)) for v in itertools.product(*param_grid.values())]
        # rmses = []  # Store the RMSEs for each params here
        #
        # # Use cross validation to evaluate all parameters
        # for params in all_params:
        #     m = Prophet(**params)
        #     [m.add_regressor(col) for col in extra_regressors.columns] if extra_regressors is not None else None
        #     with self.suppress_stdout_stderr():
        #         m.fit(input_data)  # Fit model with given params
        #         df_cv = cross_validation(m, horizon=self.prediction_lags, parallel="processes")
        #         df_p = performance_metrics(df_cv, rolling_window=1)
        #         rmses.append(df_p['rmse'].values[0])
        #
        # # Find the best parameters
        # tuning_results = pd.DataFrame(all_params)
        # tuning_results['rmse'] = rmses
        #
        # best_params = all_params[np.argmin(rmses)]
        # print(best_params)
        #
        # self.fbmodel = Prophet(**best_params)
        # [self.fbmodel.add_regressor(col) for col in extra_regressors.columns] if extra_regressors is not None else None
        # with self.suppress_stdout_stderr():
        #     self.fbmodel.fit(input_data)

    def predict(self, future_dataframe: DataFrame, extra_regressors: DataFrame = None) -> DataFrame:
        """Overrides PredictionModel.predict()"""
        future = future_dataframe.reset_index()
        future.rename(columns={'index': 'ds'}, inplace=True)

        if extra_regressors is not None:
            future.set_index('ds', inplace=True)
            future = future.join(extra_regressors.copy())
            future.reset_index(inplace=True)

        forecast = self.fbmodel.predict(future)


        forecast.set_index('ds', inplace=True)

        return forecast


class suppress_stdout_stderr(object):
    """
    A context manager for doing a "deep suppression" of stdout and stderr in
    Python, i.e. will suppress all print, even if the print originates in a
    compiled C/Fortran sub-function.
       This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).

    """

    def __init__(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)
        # Close the null files
        for fd in self.null_fds + self.save_fds:
            os.close(fd)



import logging

import dateparser
import pandas as pd
from pandas import DataFrame

log = logging.getLogger(__name__)


def ingest_timeseries(param_config: dict):
    """Retrieve the time-series data at the URL specified in `param_config['input parameters']` and return it in a
    Pandas' DataFrame.
    This can be used for the initial data ingestion, i.e. to ingest the initial time-series which will be predicted.

    Parameters
    ----------
    param_config : dict
        A dictionary corresponding to a TIMEX JSON configuration file.

    Returns
    -------
    df_ingestion : DataFrame
        Pandas DataFrame corresponding to the CSV files specified in param_config, with the various pre-processing steps
        applied. In particular, a frequency has been forced to the datetime index, NaN values have been interpolated.

    Notes
    -----
    In particular, the `input_parameters` sub-dictionary part of `param_config` will be used. In `input_parameters`, the
    following options has to be specified:

    - `source_data_url`: local or remote URL pointing to a CSV file;

    Additionally, some other parameters can be specified:

    - `index_column_name`: the name of the column to use as index for the DataFrame. If not specified the first one will
      be used. This column's values will be parsed with dateparser to obtain a DateTimeIndex;
    - `frequency`: if specified, the corresponding frequency will be imposed. Refer to
      https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases for a list of possible
      values. If not specified the frequency will be infered.
    - `columns_to_load_from_url`: comma-separated string of columns' names which will be read from the CSV file. If not
      specified, all columns will be read;
    - `add_diff_column`: comma-separated string of columns' names for which a corresponding column containing the diff
      values should be created. They will be created with the name `col_name'_diff. Note that the first row of the
      dataset will be discarded;
    - `timeseries_names`: dictionary of key-values (old_name: new_name) used to rename some columns in the CSV;
    - `dateparser_options`: dictionary of key-values which will be given to `dateparser.parse()`.

    Examples
    --------
    >>> timex_dict = {
    ...  "input_parameters": {
    ...    "source_data_url": "tests/test_datasets/covid_example_data_ingestion.csv",
    ...    "columns_to_load_from_url": "data,nuovi_positivi,terapia_intensiva",
    ...    "index_column_name": "data",
    ...    "add_diff_column": "terapia_intensiva",
    ...    "timeseries_names": {
    ...        "data": "Date",
    ...        "nuovi_positivi": "Daily cases",
    ...        "terapia_intensiva": "Total intensive care",
    ...        "terapia_intensiva_diff": "Daily intensive care",
    ...    }
    ...  }
    ...}
    >>> ingest_timeseries(timex_dict)
                    Daily cases  Total intensive care  Daily intensive care
    Date
    2020-02-25           93                    35                   9.0
    2020-02-26           78                    36                   1.0
    ...                 ...                   ...                   ...
    2021-02-10        12956                  2128                 -15.0
    2021-02-11        15146                  2126                  -2.0

    [353 rows x 3 columns]
    """
    log.info('Starting the data ingestion phase.')
    input_parameters = param_config["input_parameters"]

    source_data_url = input_parameters['source_data_url']

    try:
        columns_to_load_from_url = input_parameters["columns_to_load_from_url"]
        columns_to_read = list(columns_to_load_from_url.split(','))
        # We append [columns_to_read] to read_csv to maintain the same order of columns also in the df.
        df_ingestion = pd.read_csv(source_data_url, usecols=columns_to_read)[columns_to_read]

    except (KeyError, ValueError):
        df_ingestion = pd.read_csv(source_data_url)

    try:
        index_column_name = input_parameters["index_column_name"]
    except KeyError:
        index_column_name = df_ingestion.columns[0]

    log.debug(f"Parsing {index_column_name} as datetime column...")

    dateparser_options = {
        "settings": {
            "PREFER_DAY_OF_MONTH": "first"
        }
    }

    if "dateparser_options" in input_parameters:
        dateparser_options.update(input_parameters["dateparser_options"])

    df_ingestion[index_column_name] = df_ingestion[index_column_name].apply(
            lambda x: dateparser.parse(x, **dateparser_options)
    )

    df_ingestion.set_index(index_column_name, inplace=True, drop=True)

    log.debug(f"Removing duplicates rows from dataframe; keep the last...")
    df_ingestion = df_ingestion[~df_ingestion.index.duplicated(keep='last')]

    if not df_ingestion.index.is_monotonic_increasing:
        log.warning(f"Dataframe is not ordered. Ordering it...")
        df_ingestion = df_ingestion.sort_index()

    try:
        targets = list(input_parameters["add_diff_column"].split(','))
        log.debug(f"Adding the diff columns...")
        df_ingestion = add_diff_columns(df_ingestion, targets)
    except KeyError:
        pass

    try:
        mappings = input_parameters["timeseries_names"]
        df_ingestion.reset_index(inplace=True)
        df_ingestion.rename(columns=mappings, inplace=True)
        try:
            df_ingestion.set_index(mappings.get(index_column_name), inplace=True)
        except KeyError:
            df_ingestion.set_index(index_column_name, inplace=True)
    except KeyError:
        pass

    try:
        freq = input_parameters["frequency"]
    except KeyError:
        freq = None

    df_ingestion = add_freq(df_ingestion, freq)
    df_ingestion = df_ingestion.interpolate()

    log.info(f"Finished the data-ingestion phase. Some stats:\n"
             f"-> Number of rows: {len(df_ingestion)}\n"
             f"-> Number of columns: {len(df_ingestion.columns)}\n"
             f"-> Column names: {[*df_ingestion.columns]}\n"
             f"-> Number of missing data: {[*df_ingestion.isnull().sum()]}")

    return df_ingestion


def ingest_additional_regressors(source_data_url, param_config):
    """Create a DataFrame from the data specified at source_data_url, to be used as additional regressors.

    Parameters
    ----------
    source_data_url : string
        URL of the CSV file to retrieve. Local or remote.
    param_config : dict
        A dictionary corresponding to a TIMEX JSON configuration file.

    Returns
    -------
    df_ingestion : DataFrame
        Pandas DataFrame corresponding to the CSV files specified in source_data_url.

    See Also
    --------
    ingest_timeseries :
        The logic is very similar, but this is used only to load a time-series which will be used as additional
        regressor for a multivariate prediction model.
    """

    input_parameters = param_config["input_parameters"]

    df_ingestion = pd.read_csv(source_data_url)

    try:
        index_column_name = input_parameters["index_column_name"]
    except KeyError:
        index_column_name = df_ingestion.columns[0]

    log.debug(f"Parsing {index_column_name} as datetime column...")

    if "dateparser_options" in input_parameters:
        dateparser_options = input_parameters["dateparser_options"]
        df_ingestion[index_column_name] = df_ingestion[index_column_name].apply(
            lambda x: dateparser.parse(x, **dateparser_options)
        )
    else:
        df_ingestion[index_column_name] = df_ingestion[index_column_name].apply(
            lambda x: dateparser.parse(x)
        )

    df_ingestion.set_index(index_column_name, inplace=True, drop=True)

    log.debug(f"Removing duplicates rows from dataframe; keep the last...")
    df_ingestion = df_ingestion[~df_ingestion.index.duplicated(keep='last')]

    try:
        freq = input_parameters["frequency"]
    except KeyError:
        freq = None

    df_ingestion = add_freq(df_ingestion, freq)
    df_ingestion = df_ingestion.interpolate()

    return df_ingestion


def add_freq(df, freq=None) -> DataFrame:
    """Add a frequency to the index of df. Pandas DatetimeIndex have a `frequency` attribute; this function tries to
    assign a value to that attribute.

    If the index of df is a DatetimeIndex, then this function is guaranteed to return a DataFrame with the `frequency`
    attribute set. If it is not possible to assign the frequency, datetime may be normalized (i.e. keep only the date
    part and remove hour) in order to obtain days.

    Parameters
    ----------
    df : DataFrame
        Pandas DataFrame on which a frequency will be added.

    freq : str, optional
        If this attribute is specified, then the corresponding frequency will be forced on the DataFrame. If it is not
        specified, than, the frequency will be estimated.
        `freq` should be a so called 'offset alias'; the possible values can be found at
        https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#timeseries-offset-aliases

    Returns
    -------
    local_df : DataFrame
        df with the DatetimeIndex.freq set; if df did not have a DatetimeIndex, then df is returned unmodified.

    Examples
    --------
    >>> dates = [pd.Timestamp(datetime(year=2020, month=1, day=1, hour=10, minute=00)),
    ...          pd.Timestamp(datetime(year=2020, month=1, day=2, hour=12, minute=21)),
    ...          pd.Timestamp(datetime(year=2020, month=1, day=3, hour=13, minute=30)),
    ...          pd.Timestamp(datetime(year=2020, month=1, day=4, hour=11, minute=32))]
    >>> df = pd.DataFrame(data={"a": [0, 1, 2, 3]}, index=dates)
    >>> df
                         a
    2020-01-01 10:00:00  0
    2020-01-02 12:21:00  1
    2020-01-03 13:30:00  2
    2020-01-04 11:32:00  3

    `df` does not have a fixed frequency in its DatetimeIndex:

    >>> df.index.freq
    None

    Try to apply it:

    >>> df_with_freq = add_freq(df)
    >>> df_with_freq
                a
    2020-01-01  0
    2020-01-02  1
    2020-01-03  2
    2020-01-04  3

    Hours have been removed. Check the frequency of the index:

    >>> df_with_freq.index.freq
    <Day>
    """
    local_df = df.copy()

    # Check if df has a DatetimeIndex. If not, return without doing anything.
    try:
        i = local_df.index.freq
    except:
        return local_df

    # Df has already a freq. Don't do anything.
    if local_df.index.freq is not None:
        return local_df

    if freq is not None:
        if freq == 'D':
            local_df.index = local_df.index.normalize()

        local_df = local_df.asfreq(freq=freq)
        return local_df

    if freq is None:
        freq = pd.infer_freq(local_df.index)

        if freq is None:
            local_df.index = local_df.index.normalize()
            freq = pd.infer_freq(local_df.index)

        if freq is None:
            log.warning(f"No discernible frequency found for the dataframe.")
            freq = "D"

        local_df = local_df.asfreq(freq=freq)
        return local_df


def select_timeseries_portion(data_frame, param_config):
    """This allows the user to select only a part of a time-series DataFrame, according to some criteria (e.g. date).

    Parameters
    ----------
    data_frame : DataFrame
        Pandas DataFrame from which select a portion.
    param_config : dict
        A dictionary corresponding to a TIMEX JSON configuration file.

    Returns
    -------
    df: DataFrame
        Pandas' DataFrame after the selection phase.

    Notes
    -----
    In particular, the selection_parameters sub-dictionary part of param_config will be used. In selection_parameters,
    the following options can to be specified:

    - column_name_selection: if specified, only the rows in which the value of the column named `column_name_selection`
    is equal to `value_selection` are kept. If this is specified, also `value_selection` has to be specified.
    - init_datetime: if specified, only the rows where the Datetimeindex value is greater than `init_datetime` are kept.
    - end_datetime: if specified, only the rows where the Datetimeindex value is less than `end_datetime` are kept.

    Moreover, if `dateparser_options` is specified in `param_dict[input_parameters]', then the options will be passed to
    dateparser to parse the dates.

    Examples
    --------
    >>> ds = pd.date_range('2000-01-01', periods=7)
    >>> a = numpy.arange(30, 37)
    >>> b = numpy.array([0, 0, 0, 1, 1, 1, 0])
    >>> df = DataFrame(data={"a": a, "b": b}, index=ds)
    >>> print(df)
                 a  b
    2000-01-01  30  0
    2000-01-02  31  0
    2000-01-03  32  0
    2000-01-04  33  1
    2000-01-05  34  1
    2000-01-06  35  1
    2000-01-07  36  0

    Select using an initial time and an end time:

    >>> timex_config = {
    ...   "selection_parameters" : {
    ...     "init_datetime": '2000-01-02',
    ...     "end_datetime": '2000-01-05'
    ...   }
    ...}
    >>> selected_df = select_timeseries_portion(df, timex_config)
                 a  b
    2000-01-02  31  0
    2000-01-03  32  0
    2000-01-04  33  1
    2000-01-05  34  1

    Select using the value for a column:

    >>> timex_config = {
    ...   "selection_parameters" : {
    ...    "column_name_selection": "b",
    ...    "value_selection": 1
    ...   }
    ...}
    >>> selected_df = select_timeseries_portion(df, timex_config)
                 a  b
    2000-01-04  33  1
    2000-01-05  34  1
    2000-01-06  35  1
    """
    try:
        selection_parameters = param_config["selection_parameters"]
    except KeyError:
        log.debug(f"Selection phase not requested by user. Skip.")
        return data_frame

    try:
        input_parameters = param_config['input_parameters']
    except KeyError:
        input_parameters = {}

    log.info(f"Total amount of rows before the selection phase: {len(data_frame)}")

    if "column_name_selection" in selection_parameters and "value_selection" in selection_parameters:
        column_name = param_config['selection_parameters']['column_name_selection']
        value = param_config['selection_parameters']['value_selection']

        log.debug(f"Selection over column {column_name} with value = {value}")
        data_frame = data_frame.loc[data_frame[column_name] == value]

    if "init_datetime" in selection_parameters:
        if "dateparser_options" in input_parameters:
            dateparser_options = input_parameters["dateparser_options"]
            init_datetime = dateparser.parse(selection_parameters['init_datetime'], **dateparser_options)
        else:
            init_datetime = dateparser.parse(selection_parameters['init_datetime'])

        log.debug(f"Selection over date, keep data after {init_datetime}")
        mask = (data_frame.index.to_series() >= init_datetime)
        data_frame = data_frame.loc[mask]

    if "end_datetime" in selection_parameters:
        if "dateparser_options" in input_parameters:
            dateparser_options = input_parameters["dateparser_options"]
            end_datetime = dateparser.parse(selection_parameters['end_datetime'], **dateparser_options)
        else:
            end_datetime = dateparser.parse(selection_parameters['end_datetime'])

        log.debug(f"Selection over date, keep data before {end_datetime}")
        mask = (data_frame.index.to_series() <= end_datetime)
        data_frame = data_frame.loc[mask]

    log.info(f"Total amount of rows after the selection phase: {len(data_frame)}")
    return data_frame


def add_diff_columns(data_frame: DataFrame, column_name_target_diff: [str], group_by: str = None):
    """Function for adding a 1-step diff column computed for each column name specified in `column_name_target_diff` of
    the `data_frame`.

    The function automatically removes the first row of the data_frame since the diff value is NaN.
    If the group_by parameter is specified, the data is grouped by that sub-index and then the diff is applied.

    Parameters
    ----------
    data_frame : DataFrame
        Pandas dataframe to add the diff columns on.
    column_name_target_diff : [str]
        List columns' names for which diff columns will be computed.
    group_by : str, optional
        If specified, data is grouped by this index and then the diff is applied.

    Returns
    -------
    df: DataFrame
        Pandas dataframe with the new diff columns (name of the new columns is 'name_diff') and without the first row.

    Examples
    --------
    >>> df = DataFrame(data={"a": [1, 5, 11], "b": [4, 7, 8]})
    >>> df_with_diff_columns = add_diff_columns(df, ["a"])
    >>> df_with_diff_columns
        a  b  a_diff
    1   5  7     4.0
    2  11  8     6.0
    """
    log.info(f"Total number of rows before the add diff_columns phase: {len(data_frame)}")
    log.info(f"Total number of columns before the add diff_columns phase: {len(data_frame.columns)}")

    for target in column_name_target_diff:
        tmp = data_frame[target]
        name = target + "_diff"
        if group_by:
            data_frame[name] = tmp.groupby(group_by).diff()
        else:
            data_frame[name] = tmp.diff()

    if group_by:
        nan_rows = len(data_frame.index.get_level_values(group_by).unique())
    else:
        nan_rows = 1

    data_frame = data_frame.iloc[nan_rows:]

    log.info(f"Total number of rows after the add diff_columns phase: {len(data_frame)}")
    log.info(f"Total number of columns after the add diff_columns phase: {len(data_frame.columns)}")

    return data_frame



from math import sqrt

from pandas import DataFrame, Series
from sklearn.metrics import mean_squared_error, mean_absolute_error


class ValidationPerformance:
    """
    Class for the summary of various statistical indexes relative to the performance of a prediction model.

    Parameters
    ----------
    first_used_index, optional, default None
        Index of the first value used in the time series to generate these results. This is a convenience if you already
        know that initial index of the data this performance will refer to.

    Attributes
    ----------
    MSE : float
        Mean Squared Error. Default 0
    RMSE: float
        Root Mean Squared Error. Default 0
    MAE: float
        Mean Absolute Error. Default 0
    AM: float
        Arithmetic Mean of error. Default 0
    SD: float
        Standard deviation of error. Default 0
    """
    def __init__(self, first_used_index=None):
        self.first_used_index = first_used_index
        self.MSE = 0
        self.RMSE = 0
        self.MAE = 0
        self.AM = 0
        self.SD = 0

    def set_testing_stats(self, actual: Series, predicted: Series):
        """
        Set all the statistical indexes according to input data.

        Parameters
        ----------
        actual : Series
            Actual data stored in a Pandas Series.
        predicted : Series
            Data predicted by a model, stored in a Pandas Series.

        Examples
        --------
        >>> dates = pd.date_range('2000-01-01', periods=5)
        >>> ds = pd.DatetimeIndex(dates, freq="D")
        >>> actual = np.array([1, 1, 1, 1, 1])
        >>> predicted = np.array([3, 3, 3, 3, 3])
        >>> actual_dataframe = DataFrame(data={"a": actual}, index=ds)
        >>> predicted_dataframe = DataFrame(data={"yhat": predicted}, index=ds)

        Calculate the performances.
        >>> perf = ValidationPerformance()
        >>> perf.set_testing_stats(actual_dataframe['a'], predicted_dataframe['yhat'])

        >>> print(perf.MAE)
        2.0

        >>> print(perf.MSE)
        4.0
        """
        self.MSE = mean_squared_error(actual, predicted)
        self.MAE = mean_absolute_error(actual, predicted)
        self.RMSE = sqrt(self.MSE)
        self.AM = sum([y - yhat for y, yhat in zip(actual, predicted)]) / len(actual)
        self.SD = (actual - predicted).std(ddof=0)

    def get_dict(self) -> dict:
        """
        Return all the parameters, in a dict.

        Returns
        -------
        d : dict
            All the statistics, in a dict.

        Examples
        --------
        >>> perf = ValidationPerformance()
        >>> perf.set_testing_stats(actual_dataframe['a'], predicted_dataframe['yhat'])
        >>> perf.get_dict()
        {'first_used_index': None, 'MSE': 4.0, 'RMSE': 2.0, 'MAE': 2.0, 'AM': -2.0, 'SD': 0.0}
        """
        d = {}
        for attribute, value in self.__dict__.items():
            d[attribute] = value

        return d


import logging

from pandas import DataFrame
import numpy as np
import torch
import torch.nn as nn
from sklearn.preprocessing import MinMaxScaler

from timexseries.data_prediction import PredictionModel
log = logging.getLogger(__name__)


def split_sequences(df, n_in, n_out, n_features):
    in_out_sequence = []

    for i in range(len(df)):
        # find the end of this pattern
        end_ix = i + n_in
        out_end_ix = end_ix + n_out
        # check if we are beyond the sequence
        if out_end_ix > len(df):
            break
        # gather input and output parts of the pattern
        seq_x = []
        for j in range(i, end_ix):
            this_x = []
            for k in range(0, n_features):
                this_x.append(df.iloc[j, k])
                seq_x.append(this_x)

        seq_y = list(df.iloc[end_ix:out_end_ix, 0].values)
        in_out_sequence.append((seq_x, seq_y))
    return in_out_sequence


class LSTM(nn.Module):
    def __init__(self, input_size=1, hidden_layer_size=10):
        super().__init__()
        self.hidden_layer_size = hidden_layer_size

        self.lstm = nn.LSTM(input_size, hidden_layer_size, num_layers=1)

        self.linear = nn.Linear(hidden_layer_size, 1)

        self.hidden_cell = (torch.zeros(1, 1, self.hidden_layer_size),
                            torch.zeros(1, 1, self.hidden_layer_size))

    def forward(self, input_seq):
        lstm_out, self.hidden_cell = self.lstm(input_seq.view(len(input_seq), 1, -1), self.hidden_cell)
        predictions = self.linear(lstm_out.view(len(input_seq), -1))
        return predictions[-1]


class LSTMModel(PredictionModel):
    """LSTM prediction model."""
    def __init__(self, params: dict, transformation: str = "none"):
        super().__init__(params, name="LSTM", transformation=transformation)
        self.scalers = {}

    def train(self, input_data: DataFrame, extra_regressors: DataFrame = None):
        """Overrides PredictionModel.train()"""

        if torch.cuda.is_available():
            dev = "cuda:0"
        else:
            dev = "cpu"

        if extra_regressors is not None:
            # We could apply self.transformation also on the extra regressors.
            # From tests, it looks like it doesn't change much/it worsens the forecasts.
            input_data = input_data.join(extra_regressors)
            n_features = 1 + len(extra_regressors.columns)
            for col in extra_regressors.columns:
                self.scalers[col] = MinMaxScaler(feature_range=(-1, 1))
                extra_regressors[col] = self.scalers[col].fit_transform(extra_regressors[[col]])
        else:
            n_features = 1

        input_data.reset_index(inplace=True)
        column_indices = [0, 1]
        new_names = ['ds', 'y']
        old_names = input_data.columns[column_indices]
        input_data.rename(columns=dict(zip(old_names, new_names)), inplace=True)
        input_data.set_index('ds', inplace=True)

        self.scalers['y'] = MinMaxScaler(feature_range=(-1, 1))
        input_data['y'] = self.scalers['y'].fit_transform(input_data[['y']])

        n_steps_in, n_steps_out = round(len(input_data)/4), 1

        train_inout_seq = split_sequences(input_data, n_steps_in, n_steps_out, n_features=n_features)

        for i in range(0, len(train_inout_seq)):
            x = np.array(train_inout_seq[i][0], dtype=np.float32)
            y = np.array(train_inout_seq[i][1], dtype=np.float32)
            train_inout_seq[i] = (torch.from_numpy(x), torch.tensor(y))

        self.model = LSTM(input_size=n_features)
        self.model.to(dev)

        loss_function = nn.L1Loss()
        optimizer = torch.optim.Adam(self.model.parameters(), lr=0.01)

        epochs = 10

        for i in range(epochs):
            for seq, labels in train_inout_seq:
                seq = seq.to(dev)
                labels = labels.to(dev)
                optimizer.zero_grad()
                self.model.hidden_cell = (torch.zeros(1, 1, self.model.hidden_layer_size).to(dev),
                                          torch.zeros(1, 1, self.model.hidden_layer_size).to(dev))

                y_pred = self.model(seq)

                single_loss = loss_function(y_pred, labels)
                single_loss.backward()
                optimizer.step()

        self.values_for_prediction = train_inout_seq[-1][0]
        self.len_train_set = len(input_data)

    def predict(self, future_dataframe: DataFrame, extra_regressors: DataFrame = None) -> DataFrame:
        """Overrides PredictionModel.predict()"""
        if torch.cuda.is_available():
            dev = "cuda:0"
        else:
            dev = "cpu"

        requested_prediction = len(future_dataframe) - self.len_train_set

        if extra_regressors is not None:
            extra_regressors = extra_regressors.iloc[-requested_prediction:].copy()

            for col in extra_regressors:
                self.scalers[col] = MinMaxScaler(feature_range=(-1, 1))
                extra_regressors[col] = self.scalers[col].fit_transform(extra_regressors[[col]])

            tensors_to_append = []
            for i in range(0, requested_prediction):
                val = np.array(extra_regressors.iloc[i, :], dtype=np.float32)
                tensors_to_append.append(torch.tensor(val))

        x_input = self.values_for_prediction
        x_input = x_input.to(dev)
        self.model.eval()

        for i in range(requested_prediction):
            seq = x_input[i:]
            with torch.no_grad():
                self.model.hidden = (torch.zeros(1, 1, self.model.hidden_layer_size),
                                     torch.zeros(1, 1, self.model.hidden_layer_size))
                result = self.model(seq)
                if extra_regressors is not None:
                    result = torch.cat((result, tensors_to_append[i].to(dev)))
                x_input = torch.cat((x_input, result.view(1, -1)))


        results = x_input[-requested_prediction:]
        results = results.to("cpu")
        results = [x[0] for x in results]
        actual_predictions = self.scalers['y'].inverse_transform(np.array(results).reshape(-1, 1))
        future_dataframe.iloc[-requested_prediction:, 0] = np.array(actual_predictions).flatten()

        return future_dataframe





from pandas import DataFrame
import numpy as np

from timexseries.data_prediction import PredictionModel
import pmdarima as pm


class ARIMAModel(PredictionModel):
    """ARIMA prediction model."""
    def __init__(self, params: dict, transformation: str = "none"):
        super().__init__(params, name="ARIMA", transformation=transformation)
        self.arima = None
        self.len_train_set = 0

    def train(self, input_data: DataFrame, extra_regressor: DataFrame = None):
        # This is the code needed to check if a SARIMA is needed.
        # However, fitting a SARIMA can take a huge amount of time.
        # s = pd.Series(sm.tsa.acf(input_data, nlags=round(len(input_data) / 2 - 1), fft=True))
        #
        # z99 = 2.5758293035489004
        # len_series = len(input_data)
        # threshold99 = z99 / np.sqrt(len_series)
        #
        # peaks, _ = find_peaks(s, height=threshold99, prominence=0.01)
        # s = s[peaks].sort_values(ascending=False)
        #
        # if len(s) > 0:
        #     # Possible seasonality.
        #     self.arima = pm.auto_arima(input_data, error_action='ignore', seasonal=True, suppress_warnings=True,
        #                                m=s.index[0])
        # else:
        self.arima = pm.auto_arima(input_data, error_action='ignore', seasonal=False, suppress_warnings=True)

        self.len_train_set = len(input_data)

    def predict(self, future_dataframe: DataFrame, extra_regressor: DataFrame = None) -> DataFrame:
        """Overrides PredictionModel.predict()"""
        requested_prediction = len(future_dataframe) - self.len_train_set
        predictions = self.arima.predict_in_sample()
        predictions = np.concatenate((predictions, self.arima.predict(n_periods=requested_prediction)))

        future_dataframe.yhat = predictions

        return future_dataframe



from pandas import DataFrame


class TimeSeriesContainer:
    """
    A TimeSeriesContainer collect all the relevant information useful to characterize a single time-series coming from
    the ingested dataset.

    Parameters
    ----------
    timeseries_data : DataFrame
        Historical time-series data, in the form of a DataFrame with a index and a single data column.
    models : dict
        Dictionary of ModelResult objects, all trained on this time-series.
    xcorr : dict
        Cross-correlation between the data of this time-series and all the other ones.
    historical_prediction : dict
        The historical prediction, i.e. the predictions computed on a rolling window on the historical data.
        This is useful to verify the performances of each model not only on the very last data, but throughout the
        history of the time-series, in a cross-validation fashion. This dictionary contains one entry for each model
        tested.
    """
    def __init__(self, timeseries_data: DataFrame, models: dict, xcorr: dict, historical_prediction: dict = None):
        self.timeseries_data = timeseries_data
        self.models = models
        self.xcorr = xcorr
        self.historical_prediction = historical_prediction

    def set_historical_prediction(self, historical_prediction):
        self.historical_prediction = historical_prediction



from pandas import Series
import numpy as np
from scipy.stats import yeojohnson


class Transformation:
    """
    Super-class used to represent various types of data transformation.
    """

    def apply(self, data: Series) -> Series:
        """
        Apply the transformation on each value in a Pandas Series. Returns the transformed Series, i.e. a Series with
        transformed values.

        Note that it is not guaranteed that the dtype of the returned Series is the same of `data`.

        Parameters
        ----------
        data : Series
            Data to transform.

        Returns
        -------
        Series
            Transformed data.
        """
        pass

    def inverse(self, data: Series) -> Series:
        """
        Apply the inverse of the transformation on the values of a Pandas Series of transformed values.
        Returns the data re-transformed back to the real world.

        Any class implementing Transformation should make the `inverse` method always return a Series with the same
        shape as the one of `data`. If the function is not invertible (e.g. Log), the returning values should be
        approximated. It is assumed in the rest of TIMEX that `inverse` does not fail.

        Parameters
        ----------
        data : Series
            Data to transform.

        Returns
        -------
        Series
            Transformed data.
        """
        pass


class Log(Transformation):
    """Class corresponding to a somewhat classic logarithmic feature transformation.

    Notes
    -----
    The actual function computed by this transformation is:

    .. math::
        f(x) = sign(x) * log(|x|)

    if `x` > 1, 0 otherwise.

    Note that this way time-series which contain 0 values will have its values modified, because `inverse` will return
    1 instead of 0 when returning the transformed time-series to the real world.

    The inverse function, indeed, is:

    .. math::
        f^{-1}(x) = sign(x) * e^{abs(x)}

    LogModified should be preferred.
    """
    def apply(self, data: Series) -> Series:
        return data.apply(lambda x: np.sign(x) * np.log(abs(x)) if abs(x) > 1 else 0)

    def inverse(self, data: Series) -> Series:
        return data.apply(lambda x: np.sign(x) * np.exp(abs(x)))

    def __str__(self):
        return "Log"


class LogModified(Transformation):
    """Class corresponding to the a custom variant of logarithmic feature transformation.
    In particular, this transformation tries to overcome the traditional issues of a logarithmic transformation, i.e.
    the impossibility to work on negative data and the different behaviour on 0 < x < 1.

    Notes
    -----
    The actual function computed by this transformation is:

    .. math::
        f(x) = sign(x) * log(|x| + 1)

    The inverse, instead, is:

    .. math::
        f^{-1}(x) = sign(x) * e^{(abs(x) - sign(x))}
    """
    def apply(self, data: Series) -> Series:
        return data.apply(lambda x: np.sign(x) * np.log(abs(x) + 1))

    def inverse(self, data: Series) -> Series:
        return data.apply(lambda x: np.sign(x) * np.exp(abs(x)) - np.sign(x))

    def __str__(self):
        return "modified Log"


class Identity(Transformation):
    """Class corresponding to the identity transformation.
    This is useful because the absence of a data pre-processing transformation would be a particular case for functions
    which compute predictions; instead, using this, that case is not special anymore.

    Notes
    -----
    The actual function computed by this transformation is:

    .. math::
        f(x) = x

    The inverse, instead, is:

    .. math::
        f^{-1}(x) = x
    """
    def apply(self, data: Series) -> Series:
        return data

    def inverse(self, data: Series) -> Series:
        return data

    def __str__(self):
        return "none"


class YeoJohnson(Transformation):
    """Class corresponding to the Yeo-Johnson transformation.

    Notes
    -----
    Introduced in [^1], this transformation tries to make the input data more stable.

    Warnings
    --------
    .. warning:: Yeo-Johnson is basically broken for some series with high values.
                 Follow this issue: https://github.com/scikit-learn/scikit-learn/issues/14959
                 Until this is solved, Yeo-Johnson may not work as expected and create random crashes.

    References
    ----------
    [^1]: Yeo, I. K., & Johnson, R. A. (2000). A new family of power transformations to improve normality or symmetry.
          Biometrika, 87(4), 954-959. https://doi.org/10.1093/biomet/87.4.954
    """

    def __init__(self):
        self.lmbda = 0

    def apply(self, data: Series) -> Series:
        ind = data.index
        res, lmbda = yeojohnson(data)
        self.lmbda = lmbda
        return Series(res, index=ind)

    def inverse(self, data: Series) -> Series:
        ind = data.index
        lmbda = self.lmbda
        x_inv = np.zeros_like(data)
        pos = data >= 0

        # when x >= 0
        if abs(lmbda) < np.spacing(1.):
            x_inv[pos] = np.exp(data[pos]) - 1
        else:  # lmbda != 0
            x_inv[pos] = np.power(data[pos] * lmbda + 1, 1 / lmbda) - 1

        # when x < 0
        if abs(lmbda - 2) > np.spacing(1.):
            x_inv[~pos] = 1 - np.power(-(2 - lmbda) * data[~pos] + 1,
                                       1 / (2 - lmbda))
        else:  # lmbda == 2
            x_inv[~pos] = 1 - np.exp(-data[~pos])

        return Series(x_inv, index=ind)

    def __str__(self):
        return f"Yeo-Johnson (lambda: {round(self.lmbda, 3)})"


class Diff(Transformation):
    """Class corresponding to the differentiate transformation.
    Basically, each value at time `t` is computed as the difference between the current value and the past one.
    Applying this transformation makes the transformed Series have one less value, because the first one can not be
    computed; the value is saved in order to be able to recompute `inverse`.

    Notes
    -----
    Let `X` be the time-series and `X(t)` the value of the time-series at time `t`. This transformation changes X in Y,
    where:

    .. math::
        Y(t) = X(t) - X(t-1)
    """
    def __init__(self):
        self.first_value = 0

    def apply(self, data: Series) -> Series:
        self.first_value = data[0]
        return data.diff()[1:]

    def inverse(self, data: Series) -> Series:
        return Series(np.r_[self.first_value, data].cumsum())

    def __str__(self):
        return "differentiate (1)"


def transformation_factory(tr_class: str) -> Transformation:
    """
    Given the type of the transformation, encoded as string, return the Transformation object.

    Parameters
    ----------
    tr_class : str
        Transformation type.

    Returns
    -------
    Transformation
        Transformation object.

    Examples
    --------
    Create a Pandas Series and apply the logarithmic transformation:

    >>> x = Series([2, 3, 4, 5])
    >>> tr = transformation_factory("log")
    >>> tr_x = tr.apply(x)
    >>> tr_x
    0    0.693147
    1    1.098612
    2    1.386294
    3    1.609438
    dtype: float64

    Now, let's compute the inverse transformation which should return the data to the real world:

    >>> inv_tr_x = tr.inverse(tr_x)
    >>> inv_tr_x
    0    2.0
    1    3.0
    2    4.0
    3    5.0
    dtype: float64
    """
    if tr_class == "log":
        return Log()
    elif tr_class == "log_modified":
        return LogModified()
    elif tr_class == "none":
        return Identity()
    elif tr_class == "diff":
        return Diff()
    elif tr_class == "yeo_johnson":
        return YeoJohnson()



import logging
import os
import pickle
from functools import reduce
from typing import Tuple

import dateparser
from pandas import DataFrame

from timexseries.data_ingestion import ingest_additional_regressors
from timexseries.data_prediction import PredictionModel
from timexseries.data_prediction.models.arima_predictor import ARIMAModel
from timexseries.data_prediction.models.exponentialsmoothing_predictor import ExponentialSmoothingModel
from timexseries.data_prediction.models.lstm_predictor import LSTMModel
from timexseries.data_prediction.models.mockup_predictor import MockUpModel
# from timexseries.data_prediction.models.neuralprophet_predictor import NeuralProphetModel
from timexseries.data_prediction.models.prophet_predictor import FBProphetModel
from timexseries.data_prediction.xcorr import calc_all_xcorr
from timexseries.timeseries_container import TimeSeriesContainer

log = logging.getLogger(__name__)


def prepare_extra_regressor(container: TimeSeriesContainer, model: str) -> DataFrame:
    """
    This function receives a `timexseries.timeseries_container.TimeSeriesContainer` object, which includes the time-series
    historical data and the various predictions for the future.

    The best prediction for the model 'model' is taken and appended to the original time-series, in order to obtain a
    DataFrame with the original time series and the best possible prediction.

    The resulting DataFrame is returned.

    Parameters
    ----------
    container : TimeSeriesContainer
        `timexseries.timeseries_container.TimeSeriesContainer` from which an extra-regressor should be extracted.

    model
        The model from which get the best available prediction.

    Returns
    -------
    df : DataFrame
        DataFrame with the length of the original time-series + prediction lags.
    """
    name = container.timeseries_data.columns[0]
    best_prediction = container.models[model].best_prediction

    original_ts = container.timeseries_data
    f = best_prediction.loc[:, ['yhat']]
    f.rename(columns={'yhat': name}, inplace=True)

    best_entire_forecast = original_ts.combine_first(f)

    return best_entire_forecast


def get_best_univariate_predictions(ingested_data: DataFrame, param_config: dict, total_xcorr: dict = None) -> \
        Tuple[dict, list]:
    """
    Compute, for every column in `ingested_data` (every time-series) the best univariate prediction possible.
    This is done using the models specified in `param_config` and testing the effect of the different transformations
    specified in `param_config`. Moreover, the best feature transformation found, across the possible ones, will be
    returned.

    Parameters
    ----------
    ingested_data : DataFrame
        Initial data of the time-series.
    param_config : dict
        TIMEX configuration dictionary. In particular, the `model_parameters` sub-dictionary will be used. In
        `model_parameters` the following options has to be specified:

        - `possible_transformations`: comma-separated list of transformations keywords (e.g. "none,log_modified").
        - `main_accuracy_estimator`: error metric which will be minimized as target by the procedure. E.g. "mae".
        - `models`: comma-separated list of the models to use (e.g. "fbprophet,arima").

    total_xcorr : dict, optional, default None
        Cross-correlation dictionary computed by `calc_all_xcorr`. The cross-correlation is actually not used in this
        function, however it is used to build the returned `timexseries.timeseries_container.TimeSeriesContainer`, if given.

    Returns
    -------
    dict
        Dictionary which assigns the best transformation for every used prediction model, for every time-series.
    list
        A list of `timexseries.timeseries_container.TimeSeriesContainer` objects, one for each time-series.

    Examples
    --------
    Create some fake data:
    >>> dates = pd.date_range('2000-01-01', periods=30)  # Last index is 2000-01-30
    >>> ds = pd.DatetimeIndex(dates, freq="D")
    >>> a = np.arange(30, 60)
    >>> b = np.arange(60, 90)
    >>> timeseries_dataframe = DataFrame(data={"a": a, "b": b}, index=ds)

    And create the model configuration part of the TIMEX configuration dictionary:
    >>> param_config = {
    ...   "model_parameters": {
    ...     "models": "fbprophet",  # Model(s) which will be tested.
    ...     "possible_transformations": "none,log_modified",  # Possible feature transformation to test.
    ...     "main_accuracy_estimator": "mae",
    ...     "delta_training_percentage": 20,  # Training windows will be incremented by the 20% each step...
    ...     "test_values": 5,  # Use the last 5 values as validation set.
    ...     "prediction_lags": 7,  # Predict the next 7 points after 2000-01-30.
    ...     }
    ... }

    Now, get the univariate predictions:
    >>> best_transformations, timeseries_outputs = get_best_univariate_predictions(timeseries_dataframe, param_config)

    Let's inspect the results. `best_transformations` contains the suggested feature transformations to use:
    >>> best_transformations
    {'fbprophet': {'a': 'none', 'b': 'none'}}

    It is reasonable with this simple data that no transformation is the best transformation.
    We have the `timexseries.timeseries_container.TimeSeriesContainer` list as well:
    >>> timeseries_outputs
    [<timexseries.timeseries_container.TimeSeriesContainer at 0x7f62f45d1fa0>,
     <timexseries.timeseries_container.TimeSeriesContainer at 0x7f62d4e97cd0>]

    These are the `timexseries.timeseries_container.TimeSeriesContainer` objects, one for time-series `a` and one for `b`.
    Each one has various fields, in this case the most interesting one is `models`:
    >>> timeseries_outputs[0].models
    {'fbprophet': <timexseries.data_prediction.models.predictor.ModelResult at 0x7f62f45d1d90>}

    This is the `timexseries.data_prediction.models.predictor.ModelResult` object for FBProphet that we have just computed.
    """
    transformations_to_test = [*param_config["model_parameters"]["possible_transformations"].split(",")]
    main_accuracy_estimator = param_config["model_parameters"]["main_accuracy_estimator"]
    models = [*param_config["model_parameters"]["models"].split(",")]

    best_transformations = dict.fromkeys(models, {})
    timeseries_containers = []

    try:
        max_threads = param_config['max_threads']
    except KeyError:
        try:
            max_threads = len(os.sched_getaffinity(0))
        except:
            max_threads = 1

    columns = ingested_data.columns

    for col in columns:
        model_results = {}
        timeseries_data = ingested_data[[col]]
        xcorr = total_xcorr[col] if total_xcorr is not None else None

        for model in models:
            this_model_performances = []

            log.info(f"Using model {model}...")

            for transf in transformations_to_test:
                log.info(f"Computing univariate prediction for {col} using transformation: {transf}...")
                predictor = model_factory(model, param_config=param_config, transformation=transf)
                _result = predictor.launch_model(timeseries_data.copy(), max_threads=max_threads)

                performances = _result.results
                performances.sort(key=lambda x: getattr(x.testing_performances, main_accuracy_estimator.upper()))
                performances = getattr(performances[0].testing_performances, main_accuracy_estimator.upper())

                this_model_performances.append((_result, performances, transf))

            this_model_performances.sort(key=lambda x: x[1])
            best_tr = this_model_performances[0][2]
            [log.debug(f"Error with {t}: {e}") for t, e in zip(map(lambda x: x[2], this_model_performances),
                                                               map(lambda x: x[1], this_model_performances))]
            log.info(f"Best transformation for {col} using {model}: {best_tr}")
            best_transformations[model][col] = best_tr
            model_results[model] = this_model_performances[0][0]

        timeseries_containers.append(
            TimeSeriesContainer(timeseries_data, model_results, xcorr)
        )

    return best_transformations, timeseries_containers


def get_best_multivariate_predictions(timeseries_containers: [TimeSeriesContainer], ingested_data: DataFrame,
                                      best_transformations: dict, total_xcorr: dict, param_config: dict):
    """
    Starting from the a list of `timexseries.timeseries_container.TimeSeriesContainer`, use the available univariated
    predictions and the time-series in `ingested_data`, plus eventual user-given additional regressors to compute new
    multivariate predictions.

    These new predictions will be used only if better than the univariate ones.

    Returns the updated list of `timexseries.timeseries_container.TimeSeriesContainer`.

    Parameters
    ----------
    timeseries_containers : [TimeSeriesContainer]
        Initial `timexseries.timeseries_container.TimeSeriesContainer` list from which the computation of multivariate
        predictions start. Some univariate predictions should be already present in each object: more formally, each
        `timexseries.timeseries_container.TimeSeriesContainer` should have the `model_results` attribute.
    ingested_data : DataFrame
        Initial data of the time-series.
    best_transformations : dict
        Dictionary which assigns the best transformation for every used prediction model, for every time-series. It
        should be returned by `get_best_univariate_predictions`.
    total_xcorr : dict
        Cross-correlation dictionary computed by `timexseries.data_prediction.xcorr.calc_all_xcorr`. The cross-correlation is
        used in this function, to find, among all the time-series in `ingested_data`, additional regressors for each
        time-series, if there are some.
    param_config : dict
        TIMEX configuration dictionary. In particular, the `xcorr_parameters` sub-dictionary will be used. In
        `xcorr_parameters` the following options has to be specified if `total_xcorr` parameter is not None:

        - `xcorr_mode_target`: which cross-correlation algorithm should be used as target in evaluating useful
          additional regressors. E.g. "pearson".
        - `xcorr_extra_regressor_threshold`: the minimum absolute value of cross-correlation which indicates a useful
          extra-regressor. E.g. 0.8.

        Additionally, the `additional_regressors` part of the TIMEX configuration parameter dictionary can be used by
        the user to specify additional CSV paths to time-series data to use as extra-regressor.
        It should be a dictionary in the form "target time-series": "path of the additional extra-regressors".
        The key "_all" is a special key which indicates a path to additional extra-regressors which will be used for
        any time-series.

    Returns
    -------
    list
        A list of `timexseries.timeseries_container.TimeSeriesContainer` objects, one for each time-series.

    Examples
    --------
    We will create ad-hoc time-series in which using a multivariate model will perform better than using a univariate
    one.

    >>> dates = pd.date_range('2000-01-01', periods=30)  # Last index is 2000-01-30
    >>> ds = pd.DatetimeIndex(dates, freq="D")
    >>>
    >>> x = np.linspace(0, 2 * np.pi, 60)
    >>> y = np.sin(x)
    >>>
    >>> np.random.seed(0)
    >>> noise = np.random.normal(0, 2.0, 60)
    >>> y = y + noise
    >>>
    >>> a = y[:30]
    >>> b = y[5:35]
    >>>
    >>> timeseries_dataframe = DataFrame(data={"a": a, "b": b}, index=ds)

    In this dataset the time-series `b` can be used to better predict `a`... simply because it is the same series, but
    traslated!

    Try to perform the computations.

    >>> param_config = {
    ...     "model_parameters": {
    ...         "models": "LSTM",
    ...         "possible_transformations": "none,log_modified",
    ...         "main_accuracy_estimator": "mae",
    ...         "delta_training_percentage": 20,
    ...         "test_values": 5,
    ...         "prediction_lags": 7,
    ...     },
    ...     "xcorr_parameters": {
    ...         "xcorr_mode_target": "pearson",
    ...         "xcorr_extra_regressor_threshold": 0.7,
    ...         "xcorr_max_lags": 6,
    ...         "xcorr_mode": "pearson"
    ...     }
    ... }
    >>> xcorr = calc_all_xcorr(timeseries_dataframe, param_config)
    >>> best_transformations, timeseries_outputs = get_best_univariate_predictions(timeseries_dataframe, param_config)
    >>> timeseries_outputs = get_best_multivariate_predictions(timeseries_outputs, timeseries_dataframe,
    >>>                                                        best_transformations, xcorr, param_config)

    From the log, we can see:
    >>> "INFO:timexseries.data_prediction.pipeline:Found useful extra-regressors: Index(['b'], dtype='object'). Re-compute the prediction for a"
    >>> "INFO:timexseries.data_prediction.models.predictor:Creating a LSTM model..."
    >>> "INFO:timexseries.data_prediction.models.predictor:Model will use 5 different training sets..."
    >>> "INFO:timexseries.data_prediction.models.predictor:LSTM/NeuralProphet model. Cant use multiprocessing."
    >>> "INFO:timexseries.data_prediction.pipeline:Obtained a better error: 1.6009327718008979 vs old 1.9918351002921089"

    This means that using `b` as additional regressor for `a` made us obtain a better error.
    """
    iterations = 0
    best_forecasts_found = 0

    if total_xcorr is not None:
        xcorr_mode_target = param_config["xcorr_parameters"]["xcorr_mode_target"]
        xcorr_threshold = param_config["xcorr_parameters"]["xcorr_extra_regressor_threshold"]

    try:
        additional_regressors = param_config["additional_regressors"]
    except KeyError:
        additional_regressors = None

    models = [*param_config["model_parameters"]["models"].split(",")]

    try:
        max_threads = param_config['max_threads']
    except KeyError:
        try:
            max_threads = len(os.sched_getaffinity(0))
        except:
            max_threads = 1

    for model in models:
        log.info(f"Checking optimal predictions with model {model}")
        best_forecasts_found = 0
        iterations = 0

        while best_forecasts_found != len(ingested_data.columns):
            log.info(f"-> Found the optimal prediction for only {best_forecasts_found}")
            best_forecasts_found = 0

            for col in ingested_data.columns:
                depends_on_other_ts = False
                useful_extra_regressors = []

                log.debug(f"Look for extra regressors in other dataset's columns...")
                try:
                    local_xcorr = total_xcorr[col][xcorr_mode_target]

                    # Add extra regressors from the original dataset
                    for extra_regressor in local_xcorr.columns:
                        # Look only in correlation with future lags.
                        index_of_max = local_xcorr[extra_regressor].abs().idxmax()
                        corr = local_xcorr.loc[index_of_max, extra_regressor]
                        if abs(corr) > xcorr_threshold and index_of_max >= 0:
                            log.debug(
                                f"Found a possible extra-regressor for {col}: {extra_regressor} at lag {index_of_max}")

                            useful_extra_regressors.append(
                                prepare_extra_regressor(next(filter(
                                    lambda x: x.timeseries_data.columns[0] == extra_regressor, timeseries_containers)), model=model))
                            depends_on_other_ts = True
                    local_xcorr = total_xcorr[col]  # To give the full xcorr to Scenario
                except:
                    local_xcorr = None

                log.debug(f"Look for user-given additional regressors...")
                try:
                    additional_regressor_path = additional_regressors["_all"]
                    useful_extra_regressors.append(ingest_additional_regressors(additional_regressor_path, param_config))
                except:
                    pass

                try:
                    additional_regressor_path = additional_regressors[col]
                    useful_extra_regressors.append(ingest_additional_regressors(additional_regressor_path, param_config))
                except:
                    pass

                if len(useful_extra_regressors) == 0:
                    log.debug(f"No useful extra-regressor found for {col}: skipping...")
                    best_forecasts_found += 1
                else:
                    useful_extra_regressors = reduce(lambda x, y: x.join(y), useful_extra_regressors)
                    log.info(f"Found useful extra-regressors: {useful_extra_regressors.columns}. "
                             f"Re-compute the prediction for {col}")

                    timeseries_data = ingested_data[[col]]

                    tr = best_transformations[model][col]

                    predictor = model_factory(model, param_config, transformation=tr)
                    _result = predictor.launch_model(timeseries_data.copy(),
                                                     extra_regressors=useful_extra_regressors.copy(),
                                                     max_threads=max_threads)
                    old_this_container = next(filter(lambda x: x.timeseries_data.columns[0] == col, timeseries_containers))

                    old_errors = [x.testing_performances.MAE for x in old_this_container.models[model].results]
                    min_old_error = min(old_errors)
                    min_new_error = min([x.testing_performances.MAE for x in _result.results])

                    if min_new_error < min_old_error:
                        log.info(f"Obtained a better error: {min_new_error} vs old {min_old_error}")
                        new_model_results = old_this_container.models
                        new_model_results[model] = _result
                        new_container = TimeSeriesContainer(timeseries_data, new_model_results, local_xcorr)
                        timeseries_containers = [new_container if x.timeseries_data.columns[0] == col else x for x in timeseries_containers]
                        if not depends_on_other_ts:
                            best_forecasts_found += 1
                    else:
                        log.info(f"No improvements.")
                        best_forecasts_found += 1
            iterations += 1

    log.info(f"Found the optimal prediction for all the {best_forecasts_found} time-series in {iterations} iterations!")
    return timeseries_containers


def get_best_predictions(ingested_data: DataFrame, param_config: dict):
    """
    Starting from `ingested_data`, using the models/cross correlation settings set in `param_config`, return the best
    possible predictions in a `timexseries.timeseries_container.TimeSeriesContainer` for each time-series in `ingested_data`.

    Parameters
    ----------
    ingested_data : DataFrame
        Initial data of the time-series.

    param_config : dict
        TIMEX configuration dictionary. `get_best_univariate_predictions` and `get_best_multivariate_predictions` will
        use the various settings in `param_config`.

    Returns
    -------
    list
        A list of `timexseries.timeseries_container.TimeSeriesContainer` objects, one for each time-series.

    Examples
    --------
    This is basically the function on top of `get_best_univariate_predictions` and `get_best_multivariate_predictions`:
    it will call first the univariate and then the multivariate if the cross-correlation section is present
    in `param_config`.

    Create some data:
    >>> dates = pd.date_range('2000-01-01', periods=30)  # Last index is 2000-01-30
    >>> ds = pd.DatetimeIndex(dates, freq="D")
    >>> a = np.arange(30, 60)
    >>> b = np.arange(60, 90)
    >>>
    >>> timeseries_dataframe = DataFrame(data={"a": a, "b": b}, index=ds)

    Simply compute the predictions and get the returned `timexseries.timeseries_container.TimeSeriesContainer` objects:
    >>> timeseries_outputs = get_best_predictions(timeseries_dataframe, param_config)
    """
    if "xcorr_parameters" in param_config and len(ingested_data.columns) > 1:
        log.info(f"Computing the cross-correlation...")
        total_xcorr = calc_all_xcorr(ingested_data=ingested_data, param_config=param_config)
    else:
        total_xcorr = None

    best_transformations, timeseries_containers = get_best_univariate_predictions(ingested_data, param_config,
                                                                                  total_xcorr)

    if total_xcorr is not None or "additional_regressors" in param_config:
        timeseries_containers = get_best_multivariate_predictions(timeseries_containers=timeseries_containers, ingested_data=ingested_data,
                                                      best_transformations=best_transformations,
                                                      total_xcorr=total_xcorr,
                                                      param_config=param_config)

    return timeseries_containers


def compute_historical_predictions(ingested_data, param_config):
    """
    Compute the historical predictions, i.e. the predictions for (part) of the history of the time-series.

    Parameters
    ----------
    ingested_data : DataFrame
        Initial data of the time-series.
    param_config : dict
        TIMEX configuration dictionary. In particular, the `historical_prediction_parameters` sub-dictionary will be
        used. In `historical_prediction_parameters` the following options has to be specified:

        - `initial_index`: the point from which the historical computations will be made;
        - `save_path`: the historical computations are saved on a file, serialized with pickle. This allows the re-use
        of these predictions if TIMEX is restarted in the future.

        Additionally, the parameter `delta` can be specified: this indicates how many data points should be predicted
        every run. The default is `1`; a number greater than `1` will reduce the accuracy of the predictions because
        multiple points are predicted with the same model, but it will speed up the computation.

        `input_parameters` will be used because the `initial_index` date will be parsed with the same format provided in
        `input_parameters`, if any. Otherwise the standard `yyyy-mm-dd` format will be used.

    Returns
    -------
    list
        A list of `timexseries.timeseries_container.TimeSeriesContainer` objects, one for each time-series. These containers
        have the `historical_prediction` attribute; the predictions in `model_results` are the more recent available
        ones.

    Notes
    -----
    Historical predictions are predictions computed on past points of the time-series, but using only the data available
    until that point.

    Consider a time-series with length `p`. Consider that we want to find the historical predictions starting from the
    middle of the time-series, i.e. from index `s=p/2`.

    To do that, we take the data available from the start of the time-series to `s`, compute the prediction for the
    instant `s + 1`, and then move forward.

    When all the data has been used, we have `s` predictions, but also the real data corresponding to that predictions;
    this allows the user to check error metrics and understand the real performances of a model, on data never seen.

    These metrics can give an idea of the future performance of the model.

    Examples
    --------
    Create some sample data.
    >>> dates = pd.date_range('2000-01-01', periods=30)  # Last index is 2000-01-30
    >>> ds = pd.DatetimeIndex(dates, freq="D")
    >>> a = np.arange(30, 60)
    >>> b = np.arange(60, 90)
    >>>
    >>> timeseries_dataframe = DataFrame(data={"a": a, "b": b}, index=ds)

    Create the configuration parameters dictionary:
    >>> param_config = {
    ...     "input_parameters": {
    ...     },
    ...     "model_parameters": {
    ...         "models": "fbprophet",
    ...         "possible_transformations": "none,log_modified",
    ...         "main_accuracy_estimator": "mae",
    ...         "delta_training_percentage": 20,
    ...         "test_values": 5,
    ...         "prediction_lags": 7,
    ...     },
    ...     "historical_prediction_parameters": {
    ...         "initial_index": "2000-01-25",
    ...         "save_path": "example.pkl"
    ...     }
    ... }

    Launch the computation.
    >>> timeseries_outputs = compute_historical_predictions(timeseries_dataframe, param_config)

    Similarly to `get_best_predictions`, we have a list of `timexseries.timeseries_container.TimeSeriesContainer` objects.
    However, these objects also have an historical prediction:
    >>> timeseries_outputs[0].historical_prediction
    {'fbprophet':                  a
                 2000-01-26       55
                 2000-01-27       56
                 2000-01-28  56.3552
                 2000-01-29  58.1709
                 2000-01-30  58.9167
    }

    If multiple models were specified, `historical_prediction` dictionary would have other entries.
    """
    input_parameters = param_config["input_parameters"]
    models = [*param_config["model_parameters"]["models"].split(",")]
    save_path = param_config["historical_prediction_parameters"]["save_path"]
    try:
        hist_pred_delta = param_config["historical_prediction_parameters"]["delta"]
    except KeyError:
        hist_pred_delta = 1

    try:
        with open(save_path, 'rb') as file:
            historical_prediction = pickle.load(file)
        log.info(f"Loaded historical prediction from file...")
        current_index = historical_prediction[models[0]].index[-1]
    except FileNotFoundError:
        log.info(f"Historical prediction file not found: computing from the start...")
        starting_index = param_config["historical_prediction_parameters"]["initial_index"]

        if "dateparser_options" in input_parameters:
            dateparser_options = input_parameters["dateparser_options"]
            current_index = dateparser.parse(starting_index, **dateparser_options)
        else:
            current_index = dateparser.parse(starting_index)

        historical_prediction = {}
        for model in models:
            historical_prediction[model] = DataFrame(columns=ingested_data.columns)

    final_index = ingested_data.index[-1]
    log.info(f"Starting index: {current_index}")
    log.info(f"Final index: {final_index}")
    delta_time = 1 * ingested_data.index.freq

    if current_index == final_index:
        log.warning(f"Initial and final index are the same. I am recomputing the last point of historical prediction.")
        current_index = current_index - delta_time * hist_pred_delta

    iterations = 0
    cur = current_index
    fin = final_index

    while cur + delta_time * hist_pred_delta <= fin:
        cur += delta_time * hist_pred_delta
        iterations += 1

    additional_computation = cur != fin
    log.debug(f"Historical computations iterations: {iterations}")
    log.debug(f"Historical additional computation: {additional_computation}")

    for i in range(0, iterations):
        available_data = ingested_data[:current_index]  # Remember: this includes current_index
        log.info(f"Using data from {available_data.index[0]} to {current_index} for training...")

        timeseries_containers = get_best_predictions(available_data, param_config)

        log.info(f"Assigning the historical predictions from {current_index + delta_time} to "
                 f"{current_index + hist_pred_delta * delta_time}")
        for s in timeseries_containers:
            for model in s.models:
                p = s.models[model].best_prediction
                timeseries_name = s.timeseries_data.columns[0]
                next_preds = p.loc[current_index + delta_time:current_index + hist_pred_delta * delta_time, 'yhat']

                for index, value in next_preds.items():
                    historical_prediction[model].loc[index, timeseries_name] = value

        current_index += delta_time * hist_pred_delta

        log.info(f"Saving partial historical prediction to file...")
        with open(save_path, 'wb') as file:
            pickle.dump(historical_prediction, file, protocol=pickle.HIGHEST_PROTOCOL)

    if additional_computation:
        log.info(f"Remaining data less than requested delta time. Computing the best predictions with last data...")
        available_data = ingested_data[:current_index]  # Remember: this includes current_index
        log.info(f"Using data from {available_data.index[0]} to {current_index} for training...")

        timeseries_containers = get_best_predictions(available_data, param_config)

        log.info(f"Assigning the historical predictions from {current_index + delta_time} to "
                 f"{final_index}")
        for s in timeseries_containers:
            for model in s.models:
                p = s.models[model].best_prediction
                timeseries_name = s.timeseries_data.columns[0]
                next_preds = p.loc[current_index + delta_time:final_index, 'yhat']

                for index, value in next_preds.items():
                    historical_prediction[model].loc[index, timeseries_name] = value

        log.info(f"Saving partial historical prediction to file...")
        with open(save_path, 'wb') as file:
            pickle.dump(historical_prediction, file, protocol=pickle.HIGHEST_PROTOCOL)

    available_data = ingested_data
    timeseries_containers = get_best_predictions(available_data, param_config)

    for s in timeseries_containers:
        timeseries_name = s.timeseries_data.columns[0]
        timeseries_historical_predictions = {}
        for model in historical_prediction:
            timeseries_historical_predictions[model] = DataFrame(historical_prediction[model].loc[:, timeseries_name])
        s.set_historical_prediction(timeseries_historical_predictions)

    return timeseries_containers


def create_timeseries_containers(ingested_data: DataFrame, param_config: dict):
    """
    Entry points of the pipeline; it will compute univariate/multivariate predictions, historical predictions, or only
    create the containers with the time-series data, according to the content of `param_config`, with this logic:

    - if `historical_prediction_parameters` is in `param_config`, then `compute_historical_predictions` will be called;
    - else, if `model_parameters` is in `param_config`, then `get_best_predictions` will be called;
    - else, create a list of `timexseries.timeseries_container.TimeSeriesContainer` with only the time-series data and, if
      `xcorr_parameters` is in `param_config`, with also the cross-correlation.

    Parameters
    ----------
    ingested_data : DataFrame
        Initial data of the time-series.

    param_config : dict
        TIMEX configuration dictionary.

    Returns
    -------
    list
        A list of `timexseries.timeseries_container.TimeSeriesContainer` objects, one for each time-series.

    Examples
    --------
    The first example of `compute_historical_predictions` applies also here; calling `create_timeseries_containers` will
    produce the same identical result.

    If we remove `historical_prediction_parameters` from the `param_config`, then calling this function is the same as
    calling `get_best_predictions`.

    However, if no predictions should be made but we just want the time-series containers:
    >>> dates = pd.date_range('2000-01-01', periods=30)  # Last index is 2000-01-30
    >>> ds = pd.DatetimeIndex(dates, freq="D")
    >>> a = np.arange(30, 60)
    >>> b = np.arange(60, 90)
    >>> timeseries_dataframe = DataFrame(data={"a": a, "b": b}, index=ds)

    Create the containers:
    >>> param_config = {}
    >>> timeseries_outputs = create_timeseries_containers(timeseries_dataframe, param_config)

    Check that no models, no historical predictions and no cross-correlation are present in the containers:
    >>> print(timeseries_outputs[0].models)
    None
    >>> print(timeseries_outputs[0].historical_prediction)
    None
    >>> print(timeseries_outputs[0].xcorr)
    None

    If `xcorr_parameters` was specified, then the last command would not return None.
    Check that the time-series data is there:

    >>> print(timeseries_outputs[0].timeseries_data)
                 a
    2000-01-01  30
    2000-01-02  31
    ...
    2000-01-29  58
    2000-01-30  59
    """
    if "historical_prediction_parameters" in param_config:
        log.debug(f"Requested the computation of historical predictions.")
        timeseries_containers = compute_historical_predictions(ingested_data, param_config)
    else:
        if "model_parameters" in param_config:
            log.debug(f"Computing best predictions, without history.")
            timeseries_containers = get_best_predictions(ingested_data, param_config)
        else:
            log.debug(f"Creating containers only for data visualization.")
            timeseries_containers = []
            if "xcorr_parameters" in param_config and len(ingested_data.columns) > 1:
                total_xcorr = calc_all_xcorr(ingested_data=ingested_data, param_config=param_config)
            else:
                total_xcorr = None

            for col in ingested_data.columns:
                timeseries_data = ingested_data[[col]]
                timeseries_xcorr = total_xcorr[col] if total_xcorr is not None else None
                timeseries_containers.append(
                    TimeSeriesContainer(timeseries_data, None, timeseries_xcorr)
                )

    return timeseries_containers


def model_factory(model_class: str, param_config: dict, transformation: str = None) -> PredictionModel:
    """
    Given the name of the model, return the corresponding PredictionModel.

    Parameters
    ----------
    model_class : str
        Model type, e.g. "fbprophet"
    param_config : dict
        TIMEX configuration dictionary, to pass to the just created model.
    transformation : str, optional, default None
        Optional `transformation` parameter to pass to the just created model.

    Returns
    -------
    PredictionModel
        Prediction model of the class specified in `model_class`.

    Examples
    --------
    >>> param_config = {
    ...    "model_parameters": {
    ...        "possible_transformations": "none,log_modified",
    ...        "main_accuracy_estimator": "mae",
    ...        "delta_training_percentage": 20,
    ...        "test_values": 5,
    ...        "prediction_lags": 7,
    ...    },
    ...}

    >>> model = model_factory("fbprophet", param_config, "none")
    >>> print(type(model))
    <class 'timexseries.data_prediction.models.prophet_predictor.FBProphetModel'>
    """
    model_class = model_class.lower()
    if model_class == "fbprophet":
        return FBProphetModel(params=param_config, transformation=transformation)
    if model_class == "lstm":
        return LSTMModel(param_config, transformation)
    # if model_class == "neuralprophet":
    #     return NeuralProphetModel(param_config, transformation)
    if model_class == "mockup":
        return MockUpModel(param_config, transformation)
    if model_class == "exponentialsmoothing":
        return ExponentialSmoothingModel(param_config, transformation)
    else:
        return ARIMAModel(params=param_config, transformation=transformation)


import logging
import math
from functools import reduce
from joblib import Parallel, delayed

import pandas as pd
from pandas import DataFrame
import numpy as np

from timexseries.data_prediction.transformation import transformation_factory
from timexseries.data_prediction.validation_performances import ValidationPerformance

log = logging.getLogger(__name__)


class SingleResult:
    """
    Class for the result of a model, trained on a specific training set.

    Parameters
    ----------
    prediction : DataFrame
        Estimated prediction, using this training set
    testing_performances : ValidationPerformance
        Testing performance (`timexseries.data_prediction.validation_performances.ValidationPerformance`), on the validation
        set, obtained using this training set to train the model.
    """

    def __init__(self, prediction: DataFrame, testing_performances: ValidationPerformance):
        self.prediction = prediction
        self.testing_performances = testing_performances


class ModelResult:
    """
    Class for to collect the global results of a model trained on a time-series.

    Parameters
    ----------
    results : [SingleResult]
        List of all the results obtained using all the possible training set for this model, on the time series.
        This is useful to create plots which show how the performance vary changing the training data (e.g.
        `timexseries.data_visualization.functions.performance_plot`).
    characteristics : dict
        Model parameters. This dictionary collects human-readable characteristics of the model, e.g. the used number of
        validation points used, the length of the sliding training window, etc.
    best_prediction : DataFrame
        Prediction obtained using the best training window and _all_ the available points in the time-series. This is
        the prediction that users are most likely to want.
    """

    def __init__(self, results: [SingleResult], characteristics: dict, best_prediction: DataFrame):
        self.results = results
        self.characteristics = characteristics
        self.best_prediction = best_prediction


class PredictionModel:
    """
    Base class for every prediction model which can be used on a time series.

    Parameters
    ----------
    params : dict
        A dictionary corresponding to a TIMEX JSON configuration file.
        The various attributes, described below, will be extracted from this.
        If `params` does not contain the entry `model_parameters`, then TIMEX will attempt to load a default
        configuration parameter dictionary.
    name : str
        Class of the model.
    transformation : str, "none", optional
        The class of transformation which the model should use. If not specified, "none", corresponding to the Identity
        transformation will be used.

    Attributes
    ----------
    freq : str
        If available, the frequency of the time-series.
    test_values : int
        Number of the last points of the time-series to use for the validation set. Default -1. If this is not available
        in the configuration parameter dictionary, `test_percentage` will be used.
    test_percentage : float
        Percentage of the time-series length to used for the validation set. Default 10.
    transformation : str
        Transformation to apply to the time series before using it. Default None
    prediction_lags : int
        Number of future lags for which the prediction has to be made. Default 10.
    delta_training_percentage : float
        Length, in percentage of the time-series length, of the training windows. Default 20.
    delta_training_values : int
        Length of the training windows, obtained by computing `delta_training_percentage * length of the time-series`.
    main_accuracy_estimator : str
        Error metric to use when deciding which prediction is better. Default: "mae".
    min_values : dict
        Key-values where key is the name of a column and value is the minimum expected value in that column.
        If "_all" is in the dict the corresponding value will be used for all values in each column. Default None.
    max_values : dict
        Key-values where key is the name of a column and value is the maximum expected value in that column.
        If "_all" is in the dict the corresponding value will be used for all values in each column. Default None.
    round_to_integer : list
        List of columns name which should be rounded to integer. If "_all" is in the list, all values in every column
        will be rounded to integer. Default None.
    model_characteristics : dict
        Dictionary of values containing the main characteristics and parameters of the model. Default {}
    """
    def __init__(self, params: dict, name: str, transformation: str = "none") -> None:
        self.name = name

        log.info(f"Creating a {self.name} model...")

        # Default settings.
        self.test_percentage = 10
        self.test_values = -1  # Will be set by user, if specified a precise value, or by `launch_model()`.
        self.delta_training_percentage = 20
        self.prediction_lags = 10
        self.main_accuracy_estimator = "mae"
        self.transformation = transformation_factory(transformation)
        self.min_values = None
        self.max_values = None
        self.round_to_integer = None

        if "model_parameters" in params:
            log.debug(f"Loading user model settings...")
            model_parameters = params["model_parameters"]

            try:
                self.test_values = model_parameters["test_values"]
            except KeyError:
                pass

            try:
                self.test_percentage = model_parameters["test_percentage"]
                self.test_values = -1
            except KeyError:
                pass

            try:
                self.prediction_lags = model_parameters["prediction_lags"]
            except KeyError:
                pass

            try:
                self.delta_training_percentage = model_parameters["delta_training_percentage"]
            except KeyError:
                pass

            try:
                self.main_accuracy_estimator = model_parameters["main_accuracy_estimator"]
            except KeyError:
                pass

            try:
                self.min_values = model_parameters["min_values"]
            except KeyError:
                pass

            try:
                self.max_values = model_parameters["max_values"]
            except KeyError:
                pass

            try:
                self.round_to_integer = list(model_parameters["round_to_integer"].split(","))
            except KeyError:
                pass

        self.delta_training_values = 0
        self.model_characteristics = {}

        self.freq = ""
        log.debug(f"Finished model creation.")

    def train(self, ingested_data: DataFrame, extra_regressor: DataFrame = None):
        """
        Train the model on the first column of `ingested_data`. Additional time-series, which will be used to improve
        the training in the case of a multivariate model can be passed in the `extra_regressor` DataFrame, one for each
        column.

        The predictor, after launching `train`, is ready to make predictions for the future.

        Note that this and `predict` do not split anything in training data/validation data; this is done with other
        functions, like `launch_model`.

        Parameters
        ----------
        ingested_data : DataFrame
            Training set. The entire time-series in the first column of `ingested_data` will be used for training.

        extra_regressor : DataFrame, optional, default None
            Additional time-series to use for better predictions.

        Examples
        --------
        We will use as example the `timexseries.data_prediction.models.prophet_predictor.FBProphetModel`, an instance of
        `Predictor`.

        >>> param_config = {}  # This will make the predictor use default values...
        >>> predictor = FBProphetModel(params=param_config)

        Create some training data.

        >>> dates = pd.date_range('2000-01-01', periods=30)  # Last index is 2000-01-30
        >>> ds = pd.DatetimeIndex(dates, freq="D")
        >>> a = np.arange(30, 60)
        >>> training_dataframe = DataFrame(data={"a": a}, index=ds)

        Train the model.

        >>> predictor.train(training_dataframe)
        """
        pass

    def predict(self, future_dataframe: DataFrame, extra_regressor: DataFrame = None) -> DataFrame:
        """
        Return a DataFrame with the shape of `future_dataframe`, filled with predicted values.

        This function is used by `compute_training`; `future_dataframe` has the length of the time-series used for
        training, plus the number of desired prediction points.

        Additional `extra_regressor`, which will contain additional time-series useful to improve the prediction in case
        of multivariate models, must have the same points of `future_dataframe`.

        Parameters
        ----------
        future_dataframe : DataFrame
            DataFrame which will be filled with prediction values. This DataFrame should have the same index of the data
            used for training (plus the number of predicted values), and a column named `yhat`, which corresponds to the
            prediction that should be computed.

        extra_regressor : DataFrame, optional, default None
            DataFrame which contain optional time-series which will be used to improve the prediction. It should have
            the same shape of `future_dataframe` but with one column for each time-series. All the values should be
            present.

        Returns
        -------
        forecast : DataFrame
            DataFrame which contains the prediction computed by the model, in the column `yhat`. `forecast` may contain
            additional columns, e.g. `yhat_lower`, `yhat_upper` etc.

        Examples
        --------
        We will use as example the `timexseries.data_prediction.models.prophet_predictor.FBProphetModel`, an instance of
        `Predictor`.
        If the model has been trained, as shown in `predict` example, we can create a forecast.

        First, create the future dataframe which will be filled:

        >>> future_dates = pd.date_range('2000-01-31', periods=7)
        >>> future_ds = pd.DatetimeIndex(future_dates, freq="D")
        >>> future_df = DataFrame(columns=["yhat"], index=future_ds)
        >>> future_df
                   yhat
        ds
        2000-01-31  NaN
        2000-02-01  NaN
        2000-02-02  NaN
        2000-02-03  NaN
        2000-02-04  NaN
        2000-02-05  NaN
        2000-02-06  NaN

        Now we can create the forecast:

        >>> predictions = predictor.predict(future_dataframe=future_df)
        >>> predictions
        ds
        2000-01-31    59.992592
        2000-02-01    60.992592
        2000-02-02    61.992592
        2000-02-03    62.992592
        2000-02-04    63.992592
        2000-02-05    64.992592
        2000-02-06    65.992592
        Name: yhat, dtype: float64
        """
        pass

    def _compute_trainings(self, train_ts: DataFrame, test_ts: DataFrame, extra_regressors: DataFrame, max_threads: int):
        """
        Compute the training of a model on a set of different training sets, of increasing length.
        `train_ts` is split in `n` different training sets, according to the length of `train_ts` and the value of
        `self.delta_training_values`. The computation is split across different processes, according to the value of
        max_threads which indicates the maximum number of usable processes.

        Parameters
        ----------
        train_ts : DataFrame
            The entire training set which can be used; it will be split in different training sets, in order to test
            which sub-training-set performs better.
        test_ts : DataFrame
            Testing set to be used to compute the models' performances.
        extra_regressors : DataFrame
            Additional time-series to pass to `train` in order to improve the performances.
        max_threads : int
            Maximum number of threads to use in the training phase.

        Returns
        -------
        results : list
            List of SingleResult. Each one is the result relative to the use of a specific train set.
        """
        train_sets_number = math.ceil(len(train_ts) / self.delta_training_values)
        log.info(f"Model will use {train_sets_number} different training sets...")

        def c(targets: [int]):
            _results = []

            for _i in range(targets[0], targets[1]):
                tr = train_ts.iloc[-(_i+1) * self.delta_training_values:]

                log.debug(f"Trying with last {len(tr)} values as training set...")

                self.train(tr.copy(), extra_regressors)

                future_df = pd.DataFrame(index=pd.date_range(freq=self.freq,
                                                             start=tr.index.values[0],
                                                             periods=len(tr) + self.test_values + self.prediction_lags),
                                         columns=["yhat"], dtype=tr.iloc[:, 0].dtype)

                forecast = self.predict(future_df, extra_regressors)

                forecast.loc[:, 'yhat'] = self.transformation.inverse(forecast['yhat'])

                try:
                    forecast.loc[:, 'yhat_lower'] = self.transformation.inverse(forecast['yhat_lower'])
                    forecast.loc[:, 'yhat_upper'] = self.transformation.inverse(forecast['yhat_upper'])
                except KeyError:
                    pass

                forecast = self.adjust_forecast(train_ts.columns[0], forecast)

                testing_prediction = forecast.iloc[-self.prediction_lags - self.test_values:-self.prediction_lags]

                first_used_index = tr.index.values[0]

                tp = ValidationPerformance(first_used_index)
                tp.set_testing_stats(test_ts.iloc[:, 0], testing_prediction["yhat"])
                _results.append(SingleResult(forecast, tp))

            return _results

        # if self.name == 'NeuralProphet':
        #     log.info(f"NeuralProphet model. Cant use multiprocessing.")
        #     distributions = [[0, train_sets_number]]
        #     return c(distributions[0])

        if max_threads == 1:
            distributions = [[0, train_sets_number]]
            return c(distributions[0])
        else:
            distributions = []

            if train_sets_number % max_threads == 0:
                n_threads = max_threads
                subtraining_dim = train_sets_number // n_threads
                for i in range(0, n_threads):
                    distributions.append([i*subtraining_dim, i*subtraining_dim + subtraining_dim])
            else:
                n_threads = min(max_threads, train_sets_number)
                subtraining_dim = train_sets_number // n_threads
                for i in range(0, n_threads):
                    distributions.append([i*subtraining_dim, i*subtraining_dim+subtraining_dim])
                for k in range(0, (train_sets_number % n_threads)):
                    distributions[k][1] += 1
                    distributions[k+1::] = [[x+1, y+1] for x, y in distributions[k+1::]]

        results = Parallel(n_jobs=n_threads)(delayed(c)(distributions[i]) for i in range(0, n_threads))
        results = reduce(lambda x, y: x+y, [res for res in results])
        return results

    def _compute_best_prediction(self, ingested_data: DataFrame, training_results: [SingleResult],
                                 extra_regressors: DataFrame = None):
        """
        Given the ingested data and the training results, identify the best training window and compute a prediction
        using all the possible data, till the end of the series (hence, including the validation set).
        Parameters
        ----------
        ingested_data : DataFrame
            Initial time-series data, in a DataFrame. The first column of the DataFrame is the time-series.
        training_results : [SingleResult]
            List of `SingleResult` object: each one is the result of the model on a specific training-set.
        extra_regressors : DataFrame, optional, default None
            Additional time-series to use for better predictions.
        Returns
        -------
        DataFrame
            Best available prediction for this time-series, with this model.
        """
        training_results.sort(key=lambda x: getattr(x.testing_performances, self.main_accuracy_estimator.upper()))
        best_starting_index = training_results[0].testing_performances.first_used_index

        training_data = ingested_data.copy().loc[best_starting_index:]

        training_data.iloc[:, 0] = self.transformation.apply(training_data.iloc[:, 0])

        self.train(training_data.copy(), extra_regressors)

        future_df = pd.DataFrame(index=pd.date_range(freq=self.freq,
                                                     start=training_data.index.values[0],
                                                     periods=len(training_data) + self.prediction_lags),
                                 columns=["yhat"], dtype=training_data.iloc[:, 0].dtype)

        forecast = self.predict(future_df, extra_regressors)
        forecast.loc[:, 'yhat'] = self.transformation.inverse(forecast['yhat'])

        try:
            forecast.loc[:, 'yhat_lower'] = self.transformation.inverse(forecast['yhat_lower'])
            forecast.loc[:, 'yhat_upper'] = self.transformation.inverse(forecast['yhat_upper'])
        except KeyError:
            pass

        forecast = self.adjust_forecast(training_data.columns[0], forecast)

        return forecast

    def launch_model(self, ingested_data: DataFrame, extra_regressors: DataFrame = None, max_threads: int = 1):
        """
        Train the model on `ingested_data` and returns a `ModelResult` object.
        This function is at the highest possible level of abstraction to train a model on a time-series.

        Parameters
        ----------
        ingested_data : DataFrame
            DataFrame containing the historical time series value; it will be split in training and test parts.
        extra_regressors : DataFrame, optional, default None
            Additional time-series to passed to `train` in order to improve the performances.
        max_threads : int, optional, default 1
            Maximum number of threads to use in the training phase.

        Returns
        -------
        model_result : ModelResult
            `ModelResult` containing the results of the model, trained on ingested_data.

        Examples
        --------
        First, create some training data:

        >>> dates = pd.date_range('2000-01-01', periods=30)  # Last index is 2000-01-30
        >>> ds = pd.DatetimeIndex(dates, freq="D")
        >>> a = np.arange(30, 60)
        >>> timeseries_dataframe = DataFrame(data={"a": a}, index=ds)

        Create the model, with default values:
        >>> param_config = {}
        >>> predictor = FBProphetModel(params=param_config)

        Launch the model:
        >>> model_output = predictor.launch_model(timeseries_dataframe)

        The result is an object of class `ModelResult`: it contains the best prediction...
        >>> model_output.best_prediction['yhat']
        ds
        2000-01-22    51.0
        2000-01-23    52.0
        2000-01-24    53.0
        2000-01-25    54.0
        2000-01-26    55.0
        2000-01-27    56.0
        2000-01-28    57.0
        2000-01-29    58.0
        2000-01-30    59.0
        2000-01-31    60.0
        2000-02-01    61.0
        2000-02-02    62.0
        2000-02-03    63.0
        2000-02-04    64.0
        2000-02-05    65.0
        2000-02-06    66.0
        2000-02-07    67.0
        2000-02-08    68.0
        2000-02-09    69.0
        Name: yhat, dtype: float64

        As well as the `characteristic` dictionary, which contains useful information on the model:
        >>> model_output.characteristics
        {'name': 'FBProphet', 'delta_training_percentage': 20, 'delta_training_values': 6, 'test_values': 3,
        'transformation': <timexseries.data_prediction.transformation.Identity object at 0x7f29214b3a00>}

        The `results` attribute, instead, contains the results of the training on each of the tested sub-training-sets.
        >>> model_output.results
        [<timexseries.data_prediction.models.predictor.SingleResult at 0x7f28679e4a90>,
         <timexseries.data_prediction.models.predictor.SingleResult at 0x7f28679e48b0>,
         <timexseries.data_prediction.models.predictor.SingleResult at 0x7f286793a730>,
         <timexseries.data_prediction.models.predictor.SingleResult at 0x7f28679e4c10>,
         <timexseries.data_prediction.models.predictor.SingleResult at 0x7f2875cd2490>]

        Each `SingleResult` has an attribute `prediction`, which contains the prediction
        on the validation set (in this case, composed by the last 3 values of `timeseries_dataframe`) and
        `testing_performances` which recaps the performance, in terms of MAE, MSE, etc. of that `SingleResult` on the
        validation set.
        """
        model_characteristics = self.model_characteristics

        self.delta_training_values = int(round(len(ingested_data) * self.delta_training_percentage / 100))

        if self.test_values == -1:
            self.test_values = int(round(len(ingested_data) * (self.test_percentage / 100)))

        self.freq = pd.infer_freq(ingested_data.index)

        # We need to pass ingested data both to compute_training and compute_best_prediction, so better use copy()
        # because, otherwise, we may have side effects.
        train_ts = ingested_data.copy().iloc[:-self.test_values]
        test_ts = ingested_data.copy().iloc[-self.test_values:]

        train_ts.iloc[:, 0] = self.transformation.apply(train_ts.iloc[:, 0])

        model_training_results = self._compute_trainings(train_ts, test_ts, extra_regressors, max_threads)

        best_prediction = self._compute_best_prediction(ingested_data, model_training_results, extra_regressors)

        if extra_regressors is not None:
            model_characteristics["extra_regressors"] = ', '.join([*extra_regressors.columns])

        model_characteristics["name"] = self.name
        model_characteristics["delta_training_percentage"] = self.delta_training_percentage
        model_characteristics["delta_training_values"] = self.delta_training_values
        model_characteristics["test_values"] = self.test_values
        model_characteristics["transformation"] = self.transformation

        return ModelResult(results=model_training_results, characteristics=model_characteristics,
                           best_prediction=best_prediction)

    def adjust_forecast(self, column_name: str, df: DataFrame):
        """
        Check `s` for values below the minimum set by the user (if any) or above the maximum.
        Apply the rounding to integer, if the user specified it.

        Parameters
        ----------
        column_name : str
            Name of the column.
        df : DataFrame
            Forecast to check.

        Returns
        -------
        DataFrame
            Adjusted dataframe.
        """
        if self.min_values is not None:
            min_value = None
            if "_all" in self.min_values:
                min_value = self.min_values["_all"]
            elif column_name in self.min_values:
                min_value = self.min_values[column_name]

            if min_value is not None:
                df.loc[:, 'yhat'] = df['yhat'].apply(lambda x: min_value if x < min_value else x)
                try:
                    df.loc[:, 'yhat_lower'] = df['yhat_lower'].apply(lambda x: min_value if x < min_value else x)
                    df.loc[:, 'yhat_upper'] = df['yhat_upper'].apply(lambda x: min_value if x < min_value else x)
                except KeyError:
                    pass

        if self.max_values is not None:
            max_value = None
            if "_all" in self.max_values:
                max_value = self.max_values["_all"]
            elif column_name in self.max_values:
                max_value = self.max_values[column_name]

            if max_value is not None:
                df.loc[:, 'yhat'] = df['yhat'].apply(lambda x: max_value if x > max_value else x)
                try:
                    df.loc[:, 'yhat_lower'] = df['yhat_lower'].apply(lambda x: max_value if x > max_value else x)
                    df.loc[:, 'yhat_upper'] = df['yhat_upper'].apply(lambda x: max_value if x > max_value else x)
                except KeyError:
                    pass

        if self.round_to_integer is not None:
            if "_all" in self.round_to_integer or column_name in self.round_to_integer:
                df = df.round({"yhat": 0, "yhat_lower": 0, "yhat_upper": 0})

        return df






import logging
import warnings

from pandas import DataFrame
from sklearn.metrics import mean_squared_error, mean_absolute_error
from statsmodels.tsa.holtwinters import ExponentialSmoothing
from statsmodels.tsa.stattools import adfuller

from timexseries.data_prediction import PredictionModel
import numpy as np
import pandas as pd
import statsmodels.api as sm

log = logging.getLogger(__name__)


class ExponentialSmoothingModel(PredictionModel):
    """
    Exponential smoothing prediction model.
    This model uses the Holt Winters method and automatic parameters estimation.
    """

    def __init__(self, params: dict, transformation: str = "none"):
        super().__init__(params, name="ExponentialSmoothing", transformation=transformation)
        self.model = None
        self.len_train_set = 0

    def train(self, input_data: DataFrame, extra_regressors: DataFrame = None):
        """Overrides PredictionModel.train()"""

        def is_time_series_trend_stationary(timeseries, significance=.05):
            # Dickey-Fuller test:
            adf_test = adfuller(timeseries, autolag='AIC')

            self.pValue = adf_test[1]

            if self.pValue < significance:
                return True
            else:
                return False

        s = pd.Series(sm.tsa.acf(input_data, nlags=round(len(input_data) / 2 - 1), fft=True))
        s[0] = 0
        s[1] = 0

        z99 = 2.5758293035489004
        lenSeries = len(input_data)
        threshold99 = z99 / np.sqrt(lenSeries)

        s = s[lambda x: x > threshold99].sort_values(ascending=False)

        if len(s) > 0:
            # Possible seasonality.
            possible_sp = np.array(s.index[0:10])
            errors = pd.Series([], dtype=np.float64)
            delta_test = int(round(0.1 * len(input_data)))
            train_data = input_data[:-delta_test]
            test_data = input_data[-delta_test:]

            is_stationary = is_time_series_trend_stationary(train_data)
            loss_function = mean_absolute_error if self.main_accuracy_estimator.upper() == "MAE" else mean_squared_error

            for sp in possible_sp:
                try:
                    if is_stationary:
                        model = ExponentialSmoothing(train_data, seasonal="add", initialization_method='estimated',
                                                     seasonal_periods=sp)
                    else:
                        model = ExponentialSmoothing(train_data, trend="add", damped_trend=True, seasonal="add",
                                                     initialization_method='estimated', seasonal_periods=sp)

                    # fit model
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        model_fit = model.fit()

                    # make prediction
                    yhat_grid = model_fit.predict(start=input_data.index[0], end=input_data.index[-1])
                    errors[sp] = loss_function(test_data, yhat_grid[-delta_test:])
                except:
                    errors[sp] = np.inf

            if is_stationary:
                self.model = ExponentialSmoothing(input_data, seasonal="add", initialization_method='estimated',
                                                  seasonal_periods=errors.idxmin())
            else:
                self.model = ExponentialSmoothing(input_data, trend="add", damped_trend=True, seasonal="add",
                                                  initialization_method='estimated', seasonal_periods=errors.idxmin())
        else:
            # Don't try seasonality.
            is_stationary = is_time_series_trend_stationary(input_data)
            if is_stationary:
                self.model = ExponentialSmoothing(input_data, initialization_method='estimated')
            else:
                self.model = ExponentialSmoothing(input_data, trend="add", damped_trend=True,
                                                  initialization_method='estimated')

        self.len_train_set = len(input_data)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.model = self.model.fit()

    def predict(self, future_dataframe: DataFrame, extra_regressors: DataFrame = None) -> DataFrame:
        """Overrides PredictionModel.predict()"""
        predictions = self.model.predict(start=future_dataframe.index[0], end=future_dataframe.index[-1])

        future_dataframe.yhat = predictions

        return future_dataframe
